"""
This module provides functions for correcting the a/e energy dependence, determining the cut level and calculating survival fractions.
"""

from __future__ import annotations

import logging
import re
from datetime import datetime
from typing import Callable

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from iminuit import Minuit, cost
from matplotlib.colors import LogNorm
from scipy.stats import chi2

import pygama.math.binned_fitting as pgf
import pygama.math.histogram as pgh
import pygama.pargen.energy_cal as pgc
from pygama.math.distributions import exgauss, gauss_on_exgauss, gaussian, nb_erfc
from pygama.math.functions.gauss import nb_gauss_amp
from pygama.math.functions.hpge_peak import hpge_get_fwfm, hpge_get_fwhm, hpge_get_mode
from pygama.math.functions.sum_dists import SumDists
from pygama.pargen.survival_fractions import (
    compton_sf,
    compton_sf_sweep,
    get_sf_sweep,
    get_survival_fraction,
)
from pygama.pargen.utils import convert_to_minuit, return_nans

log = logging.getLogger(__name__)

(x_lo, x_hi, n_sig, mu, sigma, n_bkg, tau) = range(7)
par_array = [(gaussian, [mu, sigma]), (exgauss, [mu, sigma, tau])]
aoe_peak = SumDists(
    par_array,
    [n_sig, n_bkg],
    "areas",
    parameter_names=["x_lo", "x_hi", "n_sig", "mu", "sigma", "n_bkg", "tau"],
    name="aoe_peak",
)


(x_lo, x_hi, n_sig, mu, sigma, frac1, tau, n_bkg, tau_bkg) = range(9)
par_array = [
    (gauss_on_exgauss, [mu, sigma, frac1, tau]),
    (exgauss, [mu, sigma, tau_bkg]),
]
aoe_peak_with_high_tail = SumDists(
    par_array,
    [n_sig, n_bkg],
    "areas",
    parameter_names=[
        "x_lo",
        "x_hi",
        "n_sig",
        "mu",
        "sigma",
        "htail",
        "tau_sig",
        "n_bkg",
        "tau",
    ],
    name="aoe_peak_with_high_tail",
)
aoe_peak_with_high_tail.get_fwfm = hpge_get_fwfm.__get__(aoe_peak_with_high_tail)
aoe_peak_with_high_tail.get_mode = hpge_get_mode.__get__(aoe_peak_with_high_tail)
aoe_peak_with_high_tail.get_fwhm = hpge_get_fwhm.__get__(aoe_peak_with_high_tail)


def aoe_peak_guess(func, hist, bins, var, **kwargs):
    bin_centers = (bins[:-1] + bins[1:]) / 2

    mu = bin_centers[np.argmax(hist)]
    pars, _ = pgf.gauss_mode_width_max(hist, bins, var, mode_guess=mu, n_bins=5)
    bin_centres = pgh.get_bin_centers(bins)
    _, fallback_sigma, _ = pgh.get_gaussian_guess(hist, bins)

    if pars is None or (
        pars[0] > bins[-1]
        or pars[0] < bins[0]
        or pars[-1] < (0.5 * np.nanmax(hist))
        or pars[-1] > (2 * np.nanmax(hist))
        or pars[1] > 2 * fallback_sigma
        or pars[1] < 0.5 * fallback_sigma
    ):
        i_0 = np.argmax(hist)
        mu = bin_centres[i_0]
        sigma = fallback_sigma
    else:
        mu = pars[0]
        sigma = pars[1]

    ls_guess = 2 * np.sum(hist[(bin_centers > mu) & (bin_centers < (mu + 2.5 * sigma))])

    if func == aoe_peak:
        guess_dict = {
            "x_lo": bins[0],
            "x_hi": bins[-1],
            "n_sig": ls_guess,
            "mu": mu,
            "sigma": sigma,
            "n_bkg": np.sum(hist) - ls_guess,
            "tau": 0.1,
        }
        for key, guess in guess_dict.items():
            if np.isnan(guess):
                guess_dict[key] = 0

    elif func == aoe_peak_with_high_tail:
        guess_dict = {
            "x_lo": bins[0],
            "x_hi": bins[-1],
            "n_sig": ls_guess,
            "mu": mu,
            "sigma": sigma,
            "htail": 0.1,
            "tau_sig": -0.1,
            "n_bkg": np.sum(hist) - ls_guess,
            "tau": 0.1,
        }
        for key, guess in guess_dict.items():
            if np.isnan(guess):
                guess_dict[key] = 0

    elif func == exgauss:
        guess_dict = {
            "x_lo": bins[0],
            "x_hi": bins[-1],
            "area": np.sum(hist) - ls_guess,
            "mu": mu,
            "sigma": sigma,
            "tau": 0.1,
        }
        for key, guess in guess_dict.items():
            if np.isnan(guess):
                guess_dict[key] = 0

    elif func == gaussian:
        guess_dict = {
            "x_lo": bins[0],
            "x_hi": bins[-1],
            "area": ls_guess,
            "mu": mu,
            "sigma": sigma,
        }
        for key, guess in guess_dict.items():
            if np.isnan(guess):
                guess_dict[key] = 0

    for item, value in kwargs.items():
        guess_dict[item] = value

    return convert_to_minuit(guess_dict, func).values


def aoe_peak_bounds(func, guess, **kwargs):
    if func == aoe_peak:
        bounds_dict = {
            "x_lo": (None, None),
            "x_hi": (None, None),
            "n_sig": (0, None),
            "mu": ((guess["x_lo"] + guess["x_hi"]) / 2, guess["x_hi"]),
            "sigma": (
                (guess["x_hi"] - guess["x_lo"]) / 50,
                (guess["x_hi"] - guess["x_lo"]) / 4,
            ),
            "n_bkg": (0, None),
            "tau": (0, 1),
        }
    elif func == aoe_peak_with_high_tail:
        bounds_dict = {
            "x_lo": (None, None),
            "x_hi": (None, None),
            "n_sig": (0, None),
            "mu": ((guess["x_lo"] + guess["x_hi"]) / 2, guess["x_hi"]),
            "sigma": (
                (guess["x_hi"] - guess["x_lo"]) / 50,
                (guess["x_hi"] - guess["x_lo"]) / 4,
            ),
            "htail": (0, 0.5),
            "tau_sig": (None, 0),
            "n_bkg": (0, None),
            "tau": (0, 1),
        }
    elif func == exgauss:
        bounds_dict = {
            "x_lo": (None, None),
            "x_hi": (None, None),
            "area": (0, None),
            "mu": (guess["x_lo"], guess["x_hi"]),
            "sigma": (0, (guess["x_hi"] - guess["x_lo"]) / 4),
            "tau": (0, 1),
        }
    elif func == gaussian:
        bounds_dict = {
            "x_lo": (None, None),
            "x_hi": (None, None),
            "area": (0, None),
            "mu": (guess["x_lo"], guess["x_hi"]),
            "sigma": (0, (guess["x_hi"] - guess["x_lo"]) / 4),
        }

    for item, value in kwargs.items():
        bounds_dict[item] = value
    return bounds_dict


def aoe_peak_fixed(func, **kwargs):
    if func == aoe_peak:
        fixed = ["x_lo", "x_hi"]
    elif func == aoe_peak_with_high_tail:
        fixed = ["x_lo", "x_hi"]
    elif func == exgauss:
        fixed = ["x_lo", "x_hi", "mu", "sigma"]
    elif func == gaussian:
        fixed = ["x_lo", "x_hi"]
    mask = ~np.isin(func.required_args(), fixed)
    return fixed, mask


class Pol1:
    @staticmethod
    def func(x, a, b):
        return x * a + b

    @staticmethod
    def string_func(input_param):
        return f"{input_param}*a+b"

    @staticmethod
    def guess(bands, means, mean_errs):
        return [-1e-06, 5e-01]


class SigmaFit:
    @staticmethod
    def func(x, a, b, c):
        return np.where(
            (x > 0) & ((a + (b / (x + 10**-99)) ** c) > 0),
            np.sqrt(a + (b / (x + 10**-99)) ** c),
            np.nan,
        )

    @staticmethod
    def string_func(input_param):
        return f"(a+(b/({input_param}+10**-99))**c)**(0.5)"

    @staticmethod
    def guess(bands, sigmas, sigma_errs):
        return [np.nanpercentile(sigmas, 50) ** 2, 2, 2]


class SigmoidFit:
    @staticmethod
    def func(x, a, b, c, d):
        return (a + b * x) * nb_erfc(c * x + d)

    @staticmethod
    def guess(xs, ys, y_errs):
        return [np.nanmax(ys) / 2, 0, 1, 1.5]


def unbinned_aoe_fit(
    aoe: np.array,
    pdf=aoe_peak,
    display: int = 0,
) -> tuple(np.array, np.array):
    """
    Fitting function for A/E, first fits just a Gaussian before using the full pdf to fit
    if fails will return NaN values

    Parameters
    ----------
    aoe: np.array
        A/E values
    pdf: PDF
        PDF to fit to
    display: int
        Level of display

    Returns
    -------
    tuple(np.array, np.array, np.ndarray, tuple)
        Tuple of fit values, errors, covariance matrix and full fit info
    """
    if not isinstance(aoe, np.ndarray):
        aoe = np.array(aoe)

    # bin_width = (
    #     2
    #     * (np.nanpercentile(aoe, 75) - np.nanpercentile(aoe, 25))
    #     * len(aoe) ** (-1 / 3)
    # )
    # nbins = int(np.ceil((np.nanmax(aoe) - np.nanmin(aoe)) / bin_width))
    hist, bins, var = pgh.get_hist(
        aoe[(aoe < np.nanpercentile(aoe, 99)) & (aoe > np.nanpercentile(aoe, 10))],
        bins=500,
    )

    gpars = aoe_peak_guess(gaussian, hist, bins, var)
    c1_min = gpars["mu"] - 0.5 * gpars["sigma"]
    c1_max = gpars["mu"] + 3 * gpars["sigma"]
    gpars["x_lo"] = c1_min
    gpars["x_hi"] = c1_max
    # Initial fit just using Gaussian
    csig = cost.ExtendedUnbinnedNLL(
        aoe[(aoe < c1_max) & (aoe > c1_min)], gaussian.pdf_ext
    )

    msig = Minuit(csig, *gpars)

    bounds = aoe_peak_bounds(gaussian, gpars)
    for arg, val in bounds.items():
        msig.limits[arg] = val
    for fix in aoe_peak_fixed(gaussian)[0]:
        msig.fixed[fix] = True
    msig.migrad()

    # Range to fit over, below this tail behaviour more exponential, few events above
    fmin = msig.values["mu"] - 15 * msig.values["sigma"]
    if fmin < np.nanmin(aoe):
        fmin = np.nanmin(aoe)
    fmax_bkg = msig.values["mu"] - 5 * msig.values["sigma"]
    fmax = msig.values["mu"] + 5 * msig.values["sigma"]
    n_bkg_guess = len(aoe[(aoe < fmax) & (aoe > fmin)]) - msig.values["area"]

    bkg_guess = aoe_peak_guess(
        exgauss,
        hist,
        bins,
        var,
        area=n_bkg_guess,
        mu=msig.values["mu"],
        sigma=msig.values["sigma"] * 0.9,
        x_lo=fmin,
        x_hi=fmax_bkg,
    )

    cbkg = cost.ExtendedUnbinnedNLL(
        aoe[(aoe < fmax_bkg) & (aoe > fmin)], exgauss.pdf_ext
    )
    mbkg = Minuit(cbkg, *bkg_guess)

    bounds = aoe_peak_bounds(exgauss, bkg_guess)

    for arg, val in bounds.items():
        mbkg.limits[arg] = val
    for fix in aoe_peak_fixed(exgauss)[0]:
        mbkg.fixed[fix] = True
    mbkg.simplex().migrad()
    mbkg.hesse()

    nsig_guess = (
        msig.values["area"]
        - np.diff(
            exgauss.cdf_ext(
                np.array([msig.values["mu"] - 2 * msig.values["sigma"], fmax]),
                mbkg.values["area"],
                msig.values["mu"],
                msig.values["sigma"],
                mbkg.values["tau"],
            )
        )[0]
    )
    if nsig_guess < 0:
        nsig_guess = 0

    x0 = aoe_peak_guess(
        pdf,
        hist,
        bins,
        var,
        n_sig=nsig_guess,
        mu=msig.values["mu"],
        sigma=msig.values["sigma"],
        n_bkg=mbkg.values["area"],
        tau=mbkg.values["tau"],
        x_lo=fmin,
        x_hi=fmax,
    )

    bounds = aoe_peak_bounds(pdf, x0)

    gof_hist, gof_bins, gof_var = pgh.get_hist(
        aoe[(aoe < fmax) & (aoe > fmin)], bins=100, range=(fmin, fmax)
    )

    # Full fit using Gaussian signal with Gaussian tail background
    c = cost.ExtendedUnbinnedNLL(aoe[(aoe < fmax) & (aoe > fmin)], pdf.pdf_ext)
    m = Minuit(c, *x0)
    for arg, val in bounds.items():
        m.limits[arg] = val
    fixed, mask = aoe_peak_fixed(pdf)
    for fix in fixed:
        m.fixed[fix] = True
    m.migrad()
    m.hesse()

    valid1 = (
        m.valid
        & (~np.isnan(np.array(m.errors)[mask]).any())
        & (~(np.array(m.errors)[mask] == 0).all())
    )

    cs = pgf.goodness_of_fit(
        gof_hist,
        gof_bins,
        gof_var,
        pdf.get_pdf,
        m.values,
        method="Pearson",
        scale_bins=True,
    )
    cs = (cs[0], cs[1] + len(np.where(mask)[0]))

    fit1 = (m.values, m.errors, m.covariance, cs, pdf, mask, valid1, m)

    m2 = Minuit(c, *x0)
    for arg, val in bounds.items():
        m2.limits[arg] = val
    fixed, mask = aoe_peak_fixed(pdf)
    for fix in fixed:
        m2.fixed[fix] = True
    m2.simplex().migrad()
    m2.hesse()

    valid2 = (
        m2.valid
        & (~np.isnan(np.array(m2.errors)[mask]).any())
        & (~(np.array(m2.errors)[mask] == 0).all())
    )
    cs2 = pgf.goodness_of_fit(
        gof_hist,
        gof_bins,
        gof_var,
        pdf.get_pdf,
        m2.values,
        method="Pearson",
        scale_bins=True,
    )
    cs2 = (cs2[0], cs2[1] + len(np.where(mask)[0]))

    fit2 = (m2.values, m2.errors, m2.covariance, cs2, pdf, mask, valid2, m2)

    frac_errors1 = np.sum(
        np.abs(
            np.array(m.errors)[mask & (np.array(m.values) > 0)]
            / np.array(m.values)[mask & (np.array(m.values) > 0)]
        )
    )
    frac_errors2 = np.sum(
        np.abs(
            np.array(m2.errors)[mask & (np.array(m2.values) > 0)]
            / np.array(m2.values)[mask & (np.array(m2.values) > 0)]
        )
    )

    if valid2 is False:
        fit = fit1
    elif valid1 is False:
        fit = fit2
    elif cs[0] * 1.05 < cs2[0]:
        fit = fit1

    elif cs2[0] * 1.05 < cs[0]:
        fit = fit2

    elif frac_errors1 <= frac_errors2:
        fit = fit1

    elif frac_errors1 > frac_errors2:
        fit = fit2
    else:
        fit = fit1

    if display > 1:
        aoe = aoe[(aoe < fmax) & (aoe > fmin)]
        nbins = 100

        plt.figure()
        xs = np.linspace(fmin, fmax, 1000)
        counts, bins, bars = plt.hist(aoe, bins=nbins, histtype="step", label="Data")
        dx = np.diff(bins)
        plt.plot(xs, pdf.get_pdf(xs, *x0) * dx[0], label="Guess")
        plt.plot(xs, pdf.get_pdf(xs, *fit[0]) * dx[0], label="Full Fit")
        pdf.components = True
        sig, bkg = pdf.get_pdf(xs, *fit[0])
        pdf.components = False
        plt.plot(xs, sig * dx[0], label="Signal")
        plt.plot(xs, bkg * dx[0], label="Background")
        plt.plot(
            xs, gaussian.pdf_ext(xs, *msig.values)[1] * dx[0], label="Initial Gaussian"
        )
        plt.plot(xs, exgauss.pdf_ext(xs, *mbkg.values)[1] * dx[0], label="Bkg guess")
        plt.xlabel("A/E")
        plt.ylabel("Counts")
        plt.legend(loc="upper left")
        plt.show()

        plt.figure()
        bin_centers = (bins[1:] + bins[:-1]) / 2
        res = (pdf.pdf(bin_centers, *fit[0]) * dx[0]) - counts
        plt.plot(
            bin_centers,
            [re / count if count != 0 else re for re, count in zip(res, counts)],
            label="Normalised Residuals",
        )
        plt.legend(loc="upper left")
        plt.show()
        return fit[0], fit[1], fit[2], fit
    else:
        return fit[0], fit[1], fit[2], fit


def fit_time_means(tstamps, means, sigmas):
    """
    Fit the time dependence of the means of the A/E distribution

    Parameters
    ----------
    tstamps: np.array
        Timestamps of the data
    means: np.array
        Means of the A/E distribution
    sigmas: np.array
        Sigmas of the A/E distribution

    Returns: dict
        Dictionary of the time dependence of the means
    """
    out_dict = {}
    current_tstamps = []
    current_means = []
    current_sigmas = []
    rolling_mean = means[
        np.where(
            (np.abs(np.diff(means)) < (0.4 * np.array(sigmas)[1:]))
            & (~np.isnan(np.abs(np.diff(means)) < (0.4 * np.array(sigmas)[1:])))
        )[0][0]
    ]
    for i, tstamp in enumerate(tstamps):
        if (
            (
                np.abs(means[i] - rolling_mean) > 0.4 * sigmas[i]
                and np.abs(means[i] - rolling_mean) > rolling_mean * 0.01
            )
            or np.isnan(means[i])
            or np.isnan(sigmas[i])
        ):
            if i + 1 == len(means):
                out_dict[tstamp] = np.nan
            else:
                if (np.abs(means[i + 1] - means[i]) < 0.4 * sigmas[i + 1]) and not (
                    np.isnan(means[i])
                    or np.isnan(means[i + 1])
                    or np.isnan(sigmas[i])
                    or np.isnan(sigmas[i + 1])
                ):
                    for ts in current_tstamps:
                        out_dict[ts] = rolling_mean
                    rolling_mean = means[i]
                    current_means = [means[i]]
                    current_tstamps = [tstamp]
                    current_sigmas = [sigmas[i]]
                else:
                    out_dict[tstamp] = np.nan
        else:
            current_tstamps.append(tstamp)
            current_means.append(means[i])
            current_sigmas.append(sigmas[i])
            rolling_mean = np.average(
                current_means, weights=1 / np.array(current_sigmas)
            )
    for tstamp in current_tstamps:
        out_dict[tstamp] = rolling_mean
    return out_dict


def average_consecutive(tstamps, means):
    """
    Fit the time dependence of the means of the A/E distribution by average consecutive entries

    Parameters
    ----------
    tstamps: np.array
        Timestamps of the data
    means: np.array
        Means of the A/E distribution
    sigmas: np.array
        Sigmas of the A/E distribution

    Returns: dict
        Dictionary of the time dependence of the means
    """
    out_dict = {}
    for i, tstamp in enumerate(tstamps):
        if i + 1 == len(means):
            out_dict[tstamp] = means[i]
        else:
            out_dict[tstamp] = np.mean([means[i], means[i + 1]])
    return out_dict


def interpolate_consecutive(tstamps, means, times, aoe_param, output_name):
    """
    Fit the time dependence of the means of the A/E distribution by average consecutive entries

    Parameters
    ----------
    tstamps: np.array
        Timestamps of the data
    means: np.array
        Means of the A/E distribution
    sigmas: np.array
        Sigmas of the A/E distribution
    times: np.array
        Times of the mean samples in unix time format

    Returns: dict
        Dictionary of the time dependence of the means
    """
    out_dict = {}
    for i, tstamp in enumerate(tstamps):
        if i + 1 == len(means):
            out_dict[tstamp] = {
                output_name: {
                    "expression": f"{aoe_param}/a",
                    "parameters": {"a": means[i]},
                }
            }
        else:
            out_dict[tstamp] = {
                output_name: {
                    "expression": f"{aoe_param} / ((timestamp - a ) * b + c) ",
                    "parameters": {
                        "a": times[i],
                        "b": (means[i + 1] - means[i]) / (times[i + 1] - times[i]),
                        "c": means[i],
                    },
                }
            }

    return out_dict


class CalAoE:
    """
    Class for calibrating the A/E,
    this follow 5 steps:
    the time correction, drift time correction, energy correction,
    the cut level calculation and the survival fraction calculation.
    """

    def __init__(
        self,
        cal_dicts: dict = None,
        cal_energy_param: str = "cuspEmax_ctc_cal",
        eres_func: callable = lambda x: 1,
        pdf=aoe_peak,
        selection_string: str = "index==index",
        dt_corr: bool = False,
        dep_correct: bool = False,
        dt_cut: dict = None,
        dt_param: str = "dt_eff",
        high_cut_val: float = 3,
        mean_func: Callable = Pol1,
        sigma_func: Callable = SigmaFit,
        compt_bands_width: float = 20,
        debug_mode: bool = False,
    ):
        """
        Parameters
        ----------

        cal_dicts: dict
            Dictionary of calibration parameters can either be empty/None, for a single run or for multiple runs
            keyed by timestamp in the format YYYYMMDDTHHMMSSZ
        cal_energy_param: str
            Calibrated energy parameter to use for A/E calibrations and for determining peak events
        eres_func: callable
            Function to determine the energy resolution should take in a single variable the calibrated energy
        pdf: PDF
            PDF to fit to the A/E distribution
        selection_string: str
            Selection string for the data that will be passed as a query to the data dataframe
        dt_corr: bool
            Whether to correct the drift time
        dep_correct: bool
            Whether to correct the double escape peak into the single site band before cut determination
        dt_cut: dict
            Dictionary of the drift time cut parameters in the form::

                {"out_param": "dt_cut", "hard": False}

            where the out_param is the name of the parameter to cut on in the dataframe (should have been precalculated)
            and "hard" is whether to remove these events completely for survival fraction calculations
            or whether they they should only be removed in the cut determination/ A/E calibration steps
            but the survival fractions should be calculated with them included
        high_cut_val: float
            Value to cut the A/E distribution at on the high side
        mean_func: Callable
            Function to fit the energy dependence of the A/E mean, should be in the form of a class as above for Pol1
        sigma_func: Callable
            Function to fit the energy dependence of the A/E sigma, should be in the form of a class as above for Sigma_Fit
        compt_bands_width: float
            Width of the compton bands to use for the energy correction
        debug_mode: bool
            If True will raise errors if the A/E calibration fails otherwise just return NaN values

        """
        self.cal_dicts = cal_dicts if cal_dicts is not None else {}
        self.cal_energy_param = cal_energy_param
        self.eres_func = eres_func
        self.pdf = pdf
        self.selection_string = selection_string
        self.dt_corr = dt_corr
        self.dt_param = "dt_eff"
        self.dep_correct = dep_correct
        self.dt_cut = dt_cut
        if self.dt_cut is not None:
            self.dt_cut_param = dt_cut["out_param"]
            self.fit_selection = f"{self.selection_string} & {self.dt_cut_param}"
            self.dt_cut_hard = dt_cut["hard"]
        else:
            self.dt_cut_param = None
            self.dt_cut_hard = False
            self.fit_selection = self.selection_string
        self.high_cut_val = high_cut_val
        self.mean_func = mean_func
        self.sigma_func = sigma_func
        self.compt_bands_width = compt_bands_width
        self.debug_mode = debug_mode

    def update_cal_dicts(self, update_dict):
        """
        Util function for updating the calibration dictionaries
        checks if the dictionary is in the format of a single run or multiple runs
        and then updates the dictionary accordingly
        """
        if len(self.cal_dicts) > 0 and re.match(
            r"(\d{8})T(\d{6})Z", list(self.cal_dicts)[0]
        ):
            for tstamp in self.cal_dicts:
                if tstamp in update_dict:
                    self.cal_dicts[tstamp].update(update_dict[tstamp])
                else:
                    self.cal_dicts[tstamp].update(update_dict)
        else:
            self.cal_dicts.update(update_dict)

    def time_correction(
        self,
        df: pd.DataFrame,
        aoe_param: str,
        mode: str = "full",
        output_name: str = "AoE_Timecorr",
        display: int = 0,
    ):
        """
        Calculates the time correction for the A/E parameter by fitting the 1-1.3 MeV
        Compton continuum and extracting the centroid. If only a single run is passed will
        just perform a shift of the A/E parameter to 1 using the centroid otherwise for multiple
        runs the shift will be determined on the mode given in.

        Parameters
        ----------

        df: pd.DataFrame
            Dataframe containing the data
        aoe_param: str
            Name of the A/E parameter to use
        mode: str
            Mode to use for the time correction, can be "full", "partial" or "none":

            none: just use the mean of the a/e centroids to shift all the data
            partial: iterate through the centroids if vary by less than 0.4 sigma
            then group and take mean otherwise when a run higher than 0.4 sigma
            is found if it is a single run set to nan otherwise start a new block
            full : each run will be corrected individually
            average_consecutive: average the consecutive centroids
            interpolate_consecutive: interpolate between the consecutive centroids
        output_name: str
            Name of the output parameter for the time corrected A/E in the dataframe and to
            be added to the calibration dictionary
        display: int
            plot level

        """
        log.info("Starting A/E time correction")
        self.timecorr_df = pd.DataFrame()
        try:
            if "run_timestamp" in df:
                for tstamp, time_df in df.groupby("run_timestamp", sort=True):
                    try:
                        pars, errs, _, _ = unbinned_aoe_fit(
                            time_df.query(
                                f"{self.fit_selection} & ({self.cal_energy_param}>1000) & ({self.cal_energy_param}<1300)"
                            )[aoe_param],
                            pdf=self.pdf,
                            display=display,
                        )
                        self.timecorr_df = pd.concat(
                            [
                                self.timecorr_df,
                                pd.DataFrame(
                                    [
                                        {
                                            "run_timestamp": tstamp,
                                            "mean": pars["mu"],
                                            "mean_err": errs["mu"],
                                            "sigma": pars["sigma"],
                                            "sigma_err": errs["sigma"],
                                            "res": pars["sigma"] / pars["mu"],
                                            "res_err": (pars["sigma"] / pars["mu"])
                                            * np.sqrt(
                                                errs["sigma"] / pars["sigma"]
                                                + errs["mu"] / pars["mu"]
                                            ),
                                        }
                                    ]
                                ),
                            ]
                        )
                    except BaseException as e:
                        if e == KeyboardInterrupt:
                            raise (e)
                        elif self.debug_mode:
                            raise (e)
                        self.timecorr_df = pd.concat(
                            [
                                self.timecorr_df,
                                pd.DataFrame(
                                    [
                                        {
                                            "run_timestamp": tstamp,
                                            "mean": np.nan,
                                            "mean_err": np.nan,
                                            "sigma": np.nan,
                                            "sigma_err": np.nan,
                                            "res": np.nan,
                                            "res_err": np.nan,
                                        }
                                    ]
                                ),
                            ]
                        )
                self.timecorr_df.set_index("run_timestamp", inplace=True)
                if len(self.timecorr_df) > 1:
                    if mode == "partial":
                        time_dict = fit_time_means(
                            np.array(self.timecorr_df.index),
                            np.array(self.timecorr_df["mean"]),
                            np.array(self.timecorr_df["sigma"]),
                        )
                        final_time_dict = {
                            tstamp: {
                                output_name: {
                                    "expression": f"{aoe_param}/a",
                                    "parameters": {"a": t_dict},
                                }
                            }
                            for tstamp, t_dict in time_dict.items()
                        }

                    elif mode == "full":
                        time_dict = {
                            time: mean
                            for time, mean in zip(
                                np.array(self.timecorr_df.index),
                                np.array(self.timecorr_df["mean"]),
                            )
                        }
                        final_time_dict = {
                            tstamp: {
                                output_name: {
                                    "expression": f"{aoe_param}/a",
                                    "parameters": {"a": t_dict},
                                }
                            }
                            for tstamp, t_dict in time_dict.items()
                        }

                    elif mode == "none":
                        time_dict = {
                            time: np.nanmean(self.timecorr_df["mean"])
                            for time in np.array(self.timecorr_df.index)
                        }
                        final_time_dict = {
                            tstamp: {
                                output_name: {
                                    "expression": f"{aoe_param}/a",
                                    "parameters": {"a": t_dict},
                                }
                            }
                            for tstamp, t_dict in time_dict.items()
                        }

                    elif mode == "average_consecutive":
                        time_dict = average_consecutive(
                            np.array(self.timecorr_df.index),
                            np.array(self.timecorr_df["mean"]),
                        )
                        final_time_dict = {
                            tstamp: {
                                output_name: {
                                    "expression": f"{aoe_param}/a",
                                    "parameters": {"a": t_dict},
                                }
                            }
                            for tstamp, t_dict in time_dict.items()
                        }

                    elif mode == "interpolate_consecutive":
                        if "timestamp" in df:
                            times = []
                            for tstamp in np.array(self.timecorr_df.index):
                                times.append(
                                    np.nanmedian(
                                        df.query(f"run_timestamp=='{tstamp}'")[
                                            "timestamp"
                                        ]
                                    )
                                )
                            final_time_dict = interpolate_consecutive(
                                np.array(self.timecorr_df.index),
                                np.array(self.timecorr_df["mean"]),
                                times,
                                aoe_param,
                                output_name,
                            )
                            time_dict = {
                                time: mean
                                for time, mean in zip(
                                    np.array(self.timecorr_df.index),
                                    np.array(self.timecorr_df["mean"]),
                                )
                            }
                        else:
                            raise ValueError(
                                "need timestamp column in dataframe for interpolation"
                            )

                    else:
                        raise ValueError("unknown mode")

                    df[output_name] = df[aoe_param] / np.array(
                        [time_dict[tstamp] for tstamp in df["run_timestamp"]]
                    )
                    self.update_cal_dicts(final_time_dict)
                else:
                    df[output_name] = (
                        df[aoe_param] / np.array(self.timecorr_df["mean"])[0]
                    )
                    self.update_cal_dicts(
                        {
                            output_name: {
                                "expression": f"{aoe_param}/a",
                                "parameters": {
                                    "a": np.array(self.timecorr_df["mean"])[0]
                                },
                            }
                        }
                    )
                log.info("A/E time correction finished")
            else:
                try:
                    pars, errs, _, _ = unbinned_aoe_fit(
                        df.query(
                            f"{self.fit_selection} & {self.cal_energy_param}>1000 & {self.cal_energy_param}<1300"
                        )[aoe_param],
                        pdf=self.pdf,
                        display=display,
                    )
                    self.timecorr_df = pd.concat(
                        [
                            self.timecorr_df,
                            pd.DataFrame(
                                [
                                    {
                                        "run_timestamp": np.nan,
                                        "mean": pars["mu"],
                                        "mean_err": errs["mu"],
                                        "sigma": pars["sigma"],
                                        "sigma_err": errs["sigma"],
                                        "res": pars["sigma"] / pars["mu"],
                                        "res_err": (pars["sigma"] / pars["mu"])
                                        * np.sqrt(
                                            errs["sigma"] / pars["sigma"]
                                            + errs["mu"] / pars["mu"]
                                        ),
                                    }
                                ]
                            ),
                        ]
                    )
                except BaseException as e:
                    if e == KeyboardInterrupt:
                        raise (e)
                    elif self.debug_mode:
                        raise (e)

                    self.timecorr_df = pd.concat(
                        [
                            self.timecorr_df,
                            pd.DataFrame(
                                [
                                    {
                                        "run_timestamp": np.nan,
                                        "mean": np.nan,
                                        "mean_err": np.nan,
                                        "sigma": np.nan,
                                        "sigma_err": np.nan,
                                        "res": np.nan,
                                        "res_err": np.nan,
                                    }
                                ]
                            ),
                        ]
                    )
                df[output_name] = df[aoe_param] / pars["mu"]
                self.update_cal_dicts(
                    {
                        output_name: {
                            "expression": f"{aoe_param}/a",
                            "parameters": {"a": pars["mu"]},
                        }
                    }
                )
                log.info("Finished A/E time correction")
        except BaseException as e:
            if e == KeyboardInterrupt:
                raise (e)
            elif self.debug_mode:
                raise (e)
            log.error("A/E time correction failed")
            df[output_name] = df[aoe_param] / np.nan
            self.update_cal_dicts(
                {
                    output_name: {
                        "expression": f"{aoe_param}/a",
                        "parameters": {"a": np.nan},
                    }
                }
            )

    def drift_time_correction(
        self,
        data: pd.DataFrame,
        aoe_param,
        out_param="AoE_DTcorr",
        display: int = 0,
    ):
        """
        Calculates the correction needed to align the two drift time regions for ICPC detectors.
        This is done by fitting the two drift time peaks in the drift time spectrum and then
        fitting the A/E peaks in each of these regions. A simple linear correction is then applied
        to align these regions.

        Parameters
        ----------

        data: pd.DataFrame
            Dataframe containing the data
        aoe_param: str
            Name of the A/E parameter to use as starting point
        output_name: str
            Name of the output parameter for the drift time corrected A/E in the dataframe and to
            be added to the calibration dictionary
        display: int
            plot level

        """
        log.info("Starting A/E drift time correction")
        self.dt_res_dict = {}
        try:
            dep_events = data.query(
                f"{self.fit_selection}&{self.cal_energy_param}>1582&{self.cal_energy_param}<1602&{self.cal_energy_param}=={self.cal_energy_param}&{aoe_param}=={aoe_param}"
            )

            hist, bins, var = pgh.get_hist(
                dep_events[aoe_param],
                bins=500,
            )
            bin_cs = (bins[1:] + bins[:-1]) / 2
            mu = bin_cs[np.argmax(hist)]
            aoe_range = [mu * 0.9, mu * 1.1]

            dt_range = [
                np.nanpercentile(dep_events[self.dt_param], 1),
                np.nanpercentile(dep_events[self.dt_param], 99),
            ]

            self.dt_res_dict["final_selection"] = (
                f"{aoe_param}>{aoe_range[0]}&{aoe_param}<{aoe_range[1]}&{self.dt_param}>{dt_range[0]}&{self.dt_param}<{dt_range[1]}&{self.dt_param}=={self.dt_param}"
            )

            final_df = dep_events.query(self.dt_res_dict["final_selection"])

            hist, bins, var = pgh.get_hist(
                final_df[self.dt_param],
                dx=32,
                range=(
                    np.nanmin(final_df[self.dt_param]),
                    np.nanmax(final_df[self.dt_param]),
                ),
            )

            bcs = pgh.get_bin_centers(bins)
            mus = bcs[pgc.get_i_local_maxima(hist / (np.sqrt(var) + 10**-99), 2)]
            pk_pars, pk_covs = pgc.hpge_fit_energy_peak_tops(
                hist,
                bins,
                var=var,
                peak_locs=mus,
                n_to_fit=5,
            )

            mus = pk_pars[:, 0]
            sigmas = pk_pars[:, 1]
            amps = pk_pars[:, 2]

            if len(mus) > 2:
                ids = np.array(
                    sorted([np.argmax(amps), np.argmax(amps[amps != np.argmax(amps)])])
                )
            else:
                ids = np.full(len(mus), True, dtype=bool)
            mus = mus[ids]
            sigmas = sigmas[ids]
            amps = amps[ids]

            self.dt_res_dict["dt_fit"] = {"mus": mus, "sigmas": sigmas, "amps": amps}

            if len(mus) < 2:
                log.info("Only 1 drift time peak found, no correction needed")
                self.alpha = 0

            else:
                aoe_grp1 = self.dt_res_dict["aoe_grp1"] = (
                    f"{self.dt_param}>{mus[0] - 2 * sigmas[0]} & {self.dt_param}<{mus[0] + 2 * sigmas[0]}"
                )
                aoe_grp2 = self.dt_res_dict["aoe_grp2"] = (
                    f"{self.dt_param}>{mus[1] - 2 * sigmas[1]} & {self.dt_param}<{mus[1] + 2 * sigmas[1]}"
                )

                aoe_pars, aoe_errs, _, _ = unbinned_aoe_fit(
                    final_df.query(aoe_grp1)[aoe_param], pdf=self.pdf, display=display
                )

                self.dt_res_dict["aoe_fit1"] = {
                    "pars": aoe_pars.to_dict(),
                    "errs": aoe_errs.to_dict(),
                }

                aoe_pars2, aoe_errs2, _, _ = unbinned_aoe_fit(
                    final_df.query(aoe_grp2)[aoe_param], pdf=self.pdf, display=display
                )

                self.dt_res_dict["aoe_fit2"] = {
                    "pars": aoe_pars2.to_dict(),
                    "errs": aoe_errs2.to_dict(),
                }

                try:
                    self.alpha = (aoe_pars2["mu"] - aoe_pars["mu"]) / (
                        (mus[0] * aoe_pars["mu"]) - (mus[1] * aoe_pars2["mu"])
                    )
                except ZeroDivisionError:
                    self.alpha = 0
                self.dt_res_dict["alpha"] = self.alpha
                log.info(f"dtcorr successful alpha:{self.alpha}")

        except BaseException as e:
            if e == KeyboardInterrupt:
                raise (e)
            elif self.debug_mode:
                raise (e)
            log.error("Drift time correction failed")
            self.alpha = 0

        data[out_param] = data[aoe_param] * (1 + self.alpha * data[self.dt_param])
        self.update_cal_dicts(
            {
                out_param: {
                    "expression": f"{aoe_param}*(1+a*{self.dt_param})",
                    "parameters": {"a": self.alpha},
                }
            }
        )

    def energy_correction(
        self,
        data: pd.DataFrame,
        aoe_param: str,
        corrected_param: str = "AoE_Corrected",
        classifier_param: str = "AoE_Classifier",
        display: int = 0,
    ):
        """
        Calculates the corrections needed for the energy dependence of the A/E.
        Does this by fitting the compton continuum in slices and then applies fits
        to the centroid and variance.

        Parameters
        ----------

        data: pd.DataFrame
            Dataframe containing the data
        aoe_param: str
            Name of the A/E parameter to use as starting point
        corrected_param: str
            Name of the output parameter for the energy mean corrected A/E to
            be added in the dataframe and  to the calibration dictionary
        classifier_param: str
            Name of the output parameter for the full mean and sigma energy corrected A/E classifier
            to be added in the dataframe and to the calibration dictionary
        display: int
            plot level

        """

        log.info("Starting A/E energy correction")
        self.energy_corr_res_dict = {}

        compt_bands = np.arange(900, 2350, self.compt_bands_width)
        peaks = np.array(
            [1080, 1094, 1459, 1512, 1552, 1592, 1620, 1650, 1670, 1830, 2105]
        )
        allowed = np.array([], dtype=bool)
        for band in compt_bands:
            allow = True
            for peak in peaks:
                if (peak - 5) > band and (peak - 5) < (band + self.compt_bands_width):
                    allow = False
                elif (peak + 5 > band) and (peak + 5) < (band + self.compt_bands_width):
                    allow = False
            allowed = np.append(allowed, allow)
        compt_bands = compt_bands[allowed]

        self.energy_corr_fits = pd.DataFrame(
            columns=[
                "compt_bands",
                "mean",
                "mean_err",
                "sigma",
                "sigma_err",
                "ratio",
                "ratio_err",
            ],
            dtype=float,
        )
        try:
            select_df = data.query(f"{self.fit_selection} & {aoe_param}>0")

            # Fit each compton band
            for band in compt_bands:
                try:
                    pars, errs, cov, _ = unbinned_aoe_fit(
                        select_df.query(
                            f"{self.cal_energy_param}>{band}&{self.cal_energy_param}< {self.compt_bands_width+band}"
                        )[aoe_param],
                        pdf=self.pdf,
                        display=display,
                    )

                    mean, mean_err = self.pdf.get_mu(pars, cov)
                    sigma, sigma_err = self.pdf.get_fwhm(pars, cov)
                    sigma = sigma / 2.355
                    sigma_err = sigma_err / 2.355

                    self.energy_corr_fits = pd.concat(
                        [
                            self.energy_corr_fits,
                            pd.DataFrame(
                                [
                                    {
                                        "compt_bands": band
                                        + self.compt_bands_width / 2,
                                        "mean": mean,
                                        "mean_err": mean_err,
                                        "sigma": sigma,
                                        "sigma_err": sigma_err,
                                        "ratio": pars["n_sig"] / pars["n_bkg"],
                                        "ratio_err": (pars["n_sig"] / pars["n_bkg"])
                                        * np.sqrt(
                                            (errs["n_sig"] / pars["n_sig"]) ** 2
                                            + (errs["n_bkg"] / pars["n_bkg"]) ** 2
                                        ),
                                    }
                                ]
                            ),
                        ]
                    )

                except BaseException as e:
                    if e == KeyboardInterrupt:
                        raise (e)
                    elif self.debug_mode:
                        raise (e)
                    self.energy_corr_fits = pd.concat(
                        [
                            self.energy_corr_fits,
                            pd.DataFrame(
                                [
                                    {
                                        "compt_bands": band,
                                        "mean": np.nan,
                                        "mean_err": np.nan,
                                        "sigma": np.nan,
                                        "sigma_err": np.nan,
                                        "ratio": np.nan,
                                        "ratio_err": np.nan,
                                    }
                                ]
                            ),
                        ]
                    )
            self.energy_corr_fits.set_index("compt_bands", inplace=True)
            valid_fits = self.energy_corr_fits.query(
                "mean_err==mean_err&sigma_err==sigma_err & sigma_err!=0 & mean_err!=0"
            )
            self.energy_corr_res_dict["n_of_valid_fits"] = len(valid_fits)
            log.info(f"{len(valid_fits)} compton bands fit successfully")
            # Fit mus against energy
            p0_mu = self.mean_func.guess(
                valid_fits.index, valid_fits["mean"], valid_fits["mean_err"]
            )
            c_mu = cost.LeastSquares(
                valid_fits.index,
                valid_fits["mean"],
                valid_fits["mean_err"],
                self.mean_func.func,
            )
            c_mu.loss = "soft_l1"
            m_mu = Minuit(c_mu, *p0_mu)
            m_mu.simplex()
            m_mu.migrad()
            m_mu.hesse()

            mu_pars = m_mu.values
            mu_errs = m_mu.errors

            csqr_mu = np.sum(
                (
                    (
                        valid_fits["mean"]
                        - self.mean_func.func(valid_fits.index, *mu_pars)
                    )
                    ** 2
                )
                / valid_fits["mean_err"]
            )
            dof_mu = len(valid_fits["mean"]) - len(mu_pars)
            p_val_mu = chi2.sf(csqr_mu, dof_mu)
            self.mean_fit_obj = m_mu

            # Fit sigma against energy
            p0_sig = self.sigma_func.guess(
                valid_fits.index, valid_fits["sigma"], valid_fits["sigma_err"]
            )
            c_sig = cost.LeastSquares(
                valid_fits.index,
                valid_fits["sigma"],
                valid_fits["sigma_err"],
                self.sigma_func.func,
            )
            c_sig.loss = "soft_l1"
            m_sig = Minuit(c_sig, *p0_sig)
            m_sig.simplex()
            m_sig.migrad()
            m_sig.hesse()

            sig_pars = m_sig.values
            sig_errs = m_sig.errors

            csqr_sig = np.sum(
                (
                    (
                        valid_fits["sigma"]
                        - self.sigma_func.func(valid_fits.index, *sig_pars)
                    )
                    ** 2
                )
                / valid_fits["sigma_err"]
            )
            dof_sig = len(valid_fits["sigma"]) - len(sig_pars)
            p_val_sig = chi2.sf(csqr_sig, dof_sig)

            self.SigmaFit_obj = m_sig

            # Get DEP fit
            n_sigma = 4
            peak = 1592
            sigma = self.eres_func(peak) / 2.355
            emin = peak - n_sigma * sigma
            emax = peak + n_sigma * sigma
            try:
                dep_pars, dep_err, _, _ = unbinned_aoe_fit(
                    select_df.query(
                        f"{self.cal_energy_param}>{emin}&{self.cal_energy_param}<{emax}"
                    )[aoe_param],
                    pdf=self.pdf,
                    display=display,
                )
            except BaseException as e:
                if e == KeyboardInterrupt:
                    raise (e)
                elif self.debug_mode:
                    raise (e)

                dep_pars, dep_err, _ = return_nans(self.pdf)

            data[corrected_param] = data[aoe_param] / self.mean_func.func(
                data[self.cal_energy_param], *mu_pars
            )
            data[classifier_param] = (data[corrected_param] - 1) / self.sigma_func.func(
                data[self.cal_energy_param], *sig_pars
            )
            log.info("Finished A/E energy successful")
            log.info(f"mean pars are {mu_pars.to_dict()}")
            log.info(f"sigma pars are {sig_pars.to_dict()}")

        except BaseException as e:
            if e == KeyboardInterrupt:
                raise (e)
            elif self.debug_mode:
                raise (e)
            log.error("A/E energy correction failed")
            mu_pars, mu_errs, mu_cov = return_nans(self.mean_func.func)
            csqr_mu, dof_mu, p_val_mu = (np.nan, np.nan, np.nan)
            csqr_sig, dof_sig, p_val_sig = (np.nan, np.nan, np.nan)
            sig_pars, sig_errs, sig_cov = return_nans(self.sigma_func.func)
            dep_pars, dep_err, dep_cov = return_nans(self.pdf)
            data[corrected_param] = data[aoe_param] * np.nan
            data[classifier_param] = data[aoe_param] * np.nan

        self.energy_corr_res_dict["mean_fits"] = {
            "func": self.mean_func.__name__,
            "module": self.mean_func.__module__,
            "expression": self.mean_func.string_func("x"),
            "pars": mu_pars.to_dict(),
            "errs": mu_errs.to_dict(),
            "p_val_mu": p_val_mu,
            "csqr_mu": (csqr_mu, dof_mu),
        }

        self.energy_corr_res_dict["SigmaFits"] = {
            "func": self.sigma_func.__name__,
            "module": self.sigma_func.__module__,
            "expression": self.sigma_func.string_func("x"),
            "pars": sig_pars.to_dict(),
            "errs": sig_errs.to_dict(),
            "p_val_mu": p_val_sig,
            "csqr_mu": (csqr_sig, dof_sig),
        }

        self.energy_corr_res_dict["dep_fit"] = {
            "func": self.pdf.name,
            "pars": dep_pars.to_dict(),
            "errs": dep_err.to_dict(),
        }

        self.update_cal_dicts(
            {
                corrected_param: {
                    "expression": f"{aoe_param}/({self.mean_func.string_func(self.cal_energy_param)})",
                    "parameters": mu_pars.to_dict(),
                },
                f"_{classifier_param}_intermediate": {
                    "expression": f"({aoe_param})-({self.mean_func.string_func(self.cal_energy_param)})",
                    "parameters": mu_pars.to_dict(),
                },
                classifier_param: {
                    "expression": f"(_{classifier_param}_intermediate)/({self.sigma_func.string_func(self.cal_energy_param)})",
                    "parameters": sig_pars.to_dict(),
                },
            }
        )

    def get_aoe_cut_fit(
        self,
        data: pd.DataFrame,
        aoe_param: str,
        peak: float,
        ranges: tuple,
        dep_acc: float = 0.9,
        output_cut_param: str = "AoE_Low_Cut",
        display: int = 1,
    ):
        """
        Determines A/E cut by sweeping through values and for each one fitting the
        DEP to determine how many events survive.
        Fits the resulting distribution and
        interpolates to get cut value at desired DEP survival fraction (typically 90%)

        Parameters
        ----------

        data: pd.DataFrame
            Dataframe containing the data
        aoe_param: str
            Name of the A/E parameter to use as starting point
        peak : float
            Energy of the peak to use for the cut determination e.g. 1592.5
        ranges: tuple
            Tuple of the range in keV below and above the peak to use for the cut determination e.g. (20,40)
        dep_acc: float
            Desired DEP survival fraction for final cut
        output_cut_param: str
            Name of the output parameter for the events passing A/E in the dataframe and to
            be added to the calibration dictionary
        display: int
            plot level

        """

        log.info("Starting A/E low cut determination")
        self.low_cut_res_dict = {}
        self.cut_fits = pd.DataFrame()

        min_range, max_range = ranges
        erange = (peak - min_range, peak + max_range)
        try:
            select_df = data.query(
                f"{self.fit_selection}&({self.cal_energy_param} > {erange[0]}) & ({self.cal_energy_param} < {erange[1]})"
            )

            # if dep_correct is True:
            #     peak_aoe = (select_df[aoe_param] / dep_mu(select_df[self.cal_energy_param])) - 1
            #     peak_aoe = select_df[aoe_param] / sig_func(select_df[self.cal_energy_param])

            self.cut_fits, _, _ = get_sf_sweep(
                select_df[self.cal_energy_param],
                select_df[aoe_param],
                None,
                peak,
                self.eres_func(peak),
                fit_range=erange,
                data_mask=None,
                cut_range=(-8, 0),
                n_samples=40,
                mode="greater",
                debug_mode=self.debug_mode,
            )

            valid_fits = self.cut_fits.query(
                f'sf_err<{(1.5 * np.nanpercentile(self.cut_fits["sf_err"], 85))}&sf_err==sf_err'
            )

            c = cost.LeastSquares(
                valid_fits.index,
                valid_fits["sf"],
                valid_fits["sf_err"],
                SigmoidFit.func,
            )
            c.loss = "soft_l1"
            m1 = Minuit(
                c,
                *SigmoidFit.guess(
                    valid_fits.index, valid_fits["sf"], valid_fits["sf_err"]
                ),
            )
            m1.simplex().migrad()
            xs = np.arange(
                np.nanmin(valid_fits.index), np.nanmax(valid_fits.index), 0.01
            )
            p = SigmoidFit.func(xs, *m1.values)
            self.cut_fit = {
                "function": SigmoidFit.__name__,
                "pars": m1.values.to_dict(),
                "errs": m1.errors.to_dict(),
            }
            self.low_cut_val = round(xs[np.argmin(np.abs(p - (100 * dep_acc)))], 3)
            log.info(f"Cut found at {self.low_cut_val}")

            data[output_cut_param] = data[aoe_param] > self.low_cut_val
            if self.dt_cut_param is not None:
                data[output_cut_param] = data[output_cut_param] & (
                    data[self.dt_cut_param]
                )
        except BaseException as e:
            if e == KeyboardInterrupt:
                raise (e)
            elif self.debug_mode:
                raise (e)
            log.error("A/E cut determination failed")
            self.low_cut_val = np.nan
            data[output_cut_param] = False

        self.update_cal_dicts(
            {
                output_cut_param: {
                    "expression": f"({aoe_param}>a)",
                    "parameters": {"a": self.low_cut_val},
                }
            }
        )

    def calculate_survival_fractions_sweep(
        self,
        data: pd.DataFrame,
        aoe_param: str,
        peaks: list,
        fit_widths: list[tuple],
        n_samples=26,
        cut_range=(-5, 5),
        mode="greater",
    ):
        """
        Calculate survival fractions for the A/E cut for a list of peaks by sweeping through values
        of the A/E cut to show how this varies

        Parameters
        ----------

        data: pd.DataFrame
            Dataframe containing the data
        aoe_param: str
            Name of the parameter in the dataframe for the final A/E classifier
        peaks: list
            List of peaks to calculate the survival fractions for
        fit_widths: list
            List of tuples of the energy range to fit the peak in
        n_samples: int
            Number of samples to take in the sweep
        cut_range: tuple
            Range of the cut to sweep through
        mode: str
            mode to use for the cut determination, can be "greater" or "less" i.e. do we want to
            keep events with A/E greater or less than the cut value

        """
        sfs = pd.DataFrame()
        peak_dfs = {}

        for i, peak in enumerate(peaks):
            try:
                select_df = data.query(
                    f"{self.selection_string}&{aoe_param}=={aoe_param}"
                )
                fwhm = self.eres_func(peak)
                if peak == 2039:
                    emin = 2 * fwhm
                    emax = 2 * fwhm
                    peak_df = select_df.query(
                        f"({self.cal_energy_param}>{peak-emin})&({self.cal_energy_param}<{peak+emax})"
                    )

                    cut_df, sf, sf_err = compton_sf_sweep(
                        peak_df[self.cal_energy_param].to_numpy(),
                        peak_df[aoe_param].to_numpy(),
                        self.low_cut_val,
                        cut_range=cut_range,
                        n_samples=n_samples,
                        mode=mode,
                        data_mask=(
                            peak_df[self.dt_cut_param].to_numpy()
                            if self.dt_cut_param is not None
                            else None
                        ),
                    )
                    sfs = pd.concat(
                        [
                            sfs,
                            pd.DataFrame([{"peak": peak, "sf": sf, "sf_err": sf_err}]),
                        ]
                    )
                    peak_dfs[peak] = cut_df
                else:
                    emin, emax = fit_widths[i]
                    fit_range = (peak - emin, peak + emax)
                    peak_df = select_df.query(
                        f"({self.cal_energy_param}>{peak-emin})&({self.cal_energy_param}<{peak+emax})"
                    )
                    cut_df, sf, sf_err = get_sf_sweep(
                        peak_df[self.cal_energy_param].to_numpy(),
                        peak_df[aoe_param].to_numpy(),
                        self.low_cut_val,
                        peak,
                        fwhm,
                        fit_range=fit_range,
                        cut_range=cut_range,
                        n_samples=n_samples,
                        mode=mode,
                        data_mask=(
                            peak_df[self.dt_cut_param].to_numpy()
                            if self.dt_cut_param is not None
                            else None
                        ),
                        debug_mode=self.debug_mode,
                    )

                    cut_df = cut_df.query(
                        f'sf_err<5*{np.nanpercentile(cut_df["sf_err"], 50)}& sf_err==sf_err & sf<=100'
                    )

                    sfs = pd.concat(
                        [
                            sfs,
                            pd.DataFrame([{"peak": peak, "sf": sf, "sf_err": sf_err}]),
                        ]
                    )
                    peak_dfs[peak] = cut_df
                log.info(f"{peak}keV: {sf:2.1f} +/- {sf_err:2.1f} %")
            except BaseException as e:
                if e == KeyboardInterrupt:
                    raise (e)
                elif self.debug_mode:
                    raise (e)
                sfs = pd.concat(
                    [
                        sfs,
                        pd.DataFrame([{"peak": peak, "sf": np.nan, "sf_err": np.nan}]),
                    ]
                )
                log.error(
                    f"A/E Survival fraction sweep determination failed for {peak} peak"
                )
        sfs.set_index("peak", inplace=True)
        return sfs, peak_dfs

    def calculate_survival_fractions(
        self,
        data: pd.DataFrame,
        aoe_param: str,
        peaks: list,
        fit_widths: list[tuple],
        mode: str = "greater",
    ):
        """
        Calculate survival fractions for the A/E cut for a list of peaks for the final
        A/E cut value

        Parameters
        ----------

        data: pd.DataFrame
            Dataframe containing the data
        aoe_param: str
            Name of the parameter in the dataframe for the final A/E classifier
        peaks: list
            List of peaks to calculate the survival fractions for
        fit_widths: list
            List of tuples of the energy range to fit the peak in
        mode: str
            mode to use for the cut determination, can be "greater" or "less" i.e. do we want to
            keep events with A/E greater or less than the cut value

        """
        sfs = pd.DataFrame()
        for i, peak in enumerate(peaks):
            fwhm = self.eres_func(peak)
            try:
                if peak == 2039:
                    emin = 2 * fwhm
                    emax = 2 * fwhm
                    peak_df = data.query(
                        f"({self.cal_energy_param}>{peak-emin})&({self.cal_energy_param}<{peak+emax})"
                    )

                    sf_dict = compton_sf(
                        peak_df[aoe_param].to_numpy(),
                        self.low_cut_val,
                        self.high_cut_val,
                        mode=mode,
                        data_mask=(
                            peak_df[self.dt_cut_param].to_numpy()
                            if self.dt_cut_param is not None
                            else None
                        ),
                    )
                    sf = sf_dict["sf"]
                    sf_err = sf_dict["sf_err"]
                    sfs = pd.concat(
                        [
                            sfs,
                            pd.DataFrame([{"peak": peak, "sf": sf, "sf_err": sf_err}]),
                        ]
                    )
                else:
                    emin, emax = fit_widths[i]
                    fit_range = (peak - emin, peak + emax)
                    peak_df = data.query(
                        f"({self.cal_energy_param}>{peak-emin})&({self.cal_energy_param}<{peak+emax})"
                    )
                    sf, sf_err, _, _ = get_survival_fraction(
                        peak_df[self.cal_energy_param].to_numpy(),
                        peak_df[aoe_param].to_numpy(),
                        self.low_cut_val,
                        peak,
                        fwhm,
                        fit_range=fit_range,
                        mode=mode,
                        high_cut=self.high_cut_val,
                        data_mask=(
                            peak_df[self.dt_cut_param].to_numpy()
                            if self.dt_cut_param is not None
                            else None
                        ),
                    )
                    sfs = pd.concat(
                        [
                            sfs,
                            pd.DataFrame([{"peak": peak, "sf": sf, "sf_err": sf_err}]),
                        ]
                    )
                log.info(f"{peak}keV: {sf:2.1f} +/- {sf_err:2.1f} %")

            except BaseException as e:
                if e == KeyboardInterrupt:
                    raise (e)
                elif self.debug_mode:
                    raise (e)
                sfs = pd.concat(
                    [
                        sfs,
                        pd.DataFrame([{"peak": peak, "sf": np.nan, "sf_err": np.nan}]),
                    ]
                )
                log.error(f"A/E survival fraction determination failed for {peak} peak")
        sfs.set_index("peak", inplace=True)
        return sfs

    def calibrate(
        self,
        df: pd.DataFrame,
        initial_aoe_param: str,
        peaks_of_interest: list = None,
        fit_widths: list[tuple] = None,
        cut_peak_idx: int = 0,
        dep_acc: float = 0.9,
        sf_nsamples: int = 11,
        sf_cut_range: tuple = (-5, 5),
        timecorr_mode: str = "full",
    ):
        """
        Main function to run a full A/E calibration with all steps i.e. time correction, drift time correction,
        energy correction, A/E cut determination and survival fraction calculation

        Parameters
        ----------

        df: pd.DataFrame
            Dataframe containing the data
        initial_aoe_param: str
            Name of the A/E parameter in dataframe to use as starting point
        peaks_of_interest: list
            List of peaks to calculate the survival fractions for
        fit_widths: list
            List of tuples of the energy range to fit the peak in for survival fraction determination
        cut_peak_idx: int
            Index of the peak in peaks of interest to use for the cut determination
        dep_acc: float
            Desired survival fraction in the peak for final cut value
        sf_nsamples: int
            Number of samples to take in the survival fraction sweep
        sf_cut_range: tuple
            Range to use for the survival fraction sweep
        timecorr_mode: str
            Mode to use for the time correction, see time_correction function for details

        """
        if peaks_of_interest is None:
            peaks_of_interest = [1592.5, 1620.5, 2039, 2103.53, 2614.50]
        if fit_widths is None:
            fit_widths = [(40, 25), (25, 40), (0, 0), (25, 40), (50, 50)]

        self.time_correction(
            df, initial_aoe_param, mode=timecorr_mode, output_name="AoE_Timecorr"
        )

        if self.dt_corr is True:
            aoe_param = "AoE_DTcorr"
            self.drift_time_correction(df, "AoE_Timecorr", out_param=aoe_param)
        else:
            aoe_param = "AoE_Timecorr"

        self.energy_correction(
            df,
            aoe_param,
            corrected_param="AoE_Corrected",
            classifier_param="AoE_Classifier",
        )

        self.get_aoe_cut_fit(
            df,
            "AoE_Classifier",
            peaks_of_interest[cut_peak_idx],
            fit_widths[cut_peak_idx],
            dep_acc,
            output_cut_param="AoE_Low_Cut",
        )

        df["AoE_Double_Sided_Cut"] = df["AoE_Low_Cut"] & (
            df["AoE_Classifier"] < self.high_cut_val
        )

        self.update_cal_dicts(
            {
                "AoE_High_Side_Cut": {
                    "expression": "(a>AoE_Classifier)",
                    "parameters": {"a": self.high_cut_val},
                }
            }
        )

        self.update_cal_dicts(
            {
                "AoE_Double_Sided_Cut": {
                    "expression": "(AoE_High_Side_Cut) & (AoE_Low_Cut)",
                    "parameters": {},
                }
            }
        )

        log.info("Compute low side survival fractions: ")
        (
            self.low_side_sfs,
            self.low_side_peak_dfs,
        ) = self.calculate_survival_fractions_sweep(
            df,
            "AoE_Classifier",
            peaks_of_interest,
            fit_widths,
            n_samples=sf_nsamples,
            cut_range=sf_cut_range,
            mode="greater",
        )

        log.info("Compute 2 side survival fractions: ")
        self.two_side_sfs = self.calculate_survival_fractions(
            df, "AoE_Classifier", peaks_of_interest, fit_widths, mode="greater"
        )

        if re.match(r"(\d{8})T(\d{6})Z", list(self.cal_dicts)[0]):
            self.low_side_sfs_by_run = {}
            self.two_side_sfs_by_run = {}
            for tstamp in self.cal_dicts:
                log.info(f"Compute survival fractions for {tstamp}: ")
                self.low_side_sfs_by_run[tstamp] = self.calculate_survival_fractions(
                    df.query(f"run_timestamp == '{tstamp}'"),
                    "AoE_Classifier",
                    peaks_of_interest,
                    fit_widths,
                    mode="greater",
                )

                self.two_side_sfs_by_run[tstamp] = self.calculate_survival_fractions(
                    df.query(f"run_timestamp == '{tstamp}'"),
                    "AoE_Classifier",
                    peaks_of_interest,
                    fit_widths,
                    mode="greater",
                )
        else:
            self.low_side_sfs_by_run = None
            self.two_side_sfs_by_run = None


# below are a few plotting functions that can be used to plot the results of the A/E calibration


def plot_aoe_mean_time(
    aoe_class, data, time_param="AoE_Timecorr", figsize=(12, 8), fontsize=12
):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize
    fig, ax = plt.subplots(1, 1)
    try:
        ax.errorbar(
            [
                datetime.strptime(tstamp, "%Y%m%dT%H%M%SZ")
                for tstamp in aoe_class.timecorr_df.index
            ],
            aoe_class.timecorr_df["mean"],
            yerr=aoe_class.timecorr_df["mean_err"],
            linestyle=" ",
            marker="x",
        )

        grouped_means = [
            cal_dict[time_param]["parameters"]["a"]
            for tstamp, cal_dict in aoe_class.cal_dicts.items()
        ]
        ax.step(
            [
                datetime.strptime(tstamp, "%Y%m%dT%H%M%SZ")
                for tstamp in aoe_class.cal_dicts
            ],
            grouped_means,
            where="post",
        )
        ax.fill_between(
            [
                datetime.strptime(tstamp, "%Y%m%dT%H%M%SZ")
                for tstamp in aoe_class.cal_dicts
            ],
            y1=np.array(grouped_means) - 0.2 * np.array(aoe_class.timecorr_df["sigma"]),
            y2=np.array(grouped_means) + 0.2 * np.array(aoe_class.timecorr_df["sigma"]),
            color="green",
            alpha=0.2,
        )
        ax.fill_between(
            [
                datetime.strptime(tstamp, "%Y%m%dT%H%M%SZ")
                for tstamp in aoe_class.cal_dicts
            ],
            y1=np.array(grouped_means) - 0.4 * np.array(aoe_class.timecorr_df["sigma"]),
            y2=np.array(grouped_means) + 0.4 * np.array(aoe_class.timecorr_df["sigma"]),
            color="yellow",
            alpha=0.2,
        )
    except Exception:
        pass
    ax.set_xlabel("time")
    ax.set_ylabel("A/E mean")
    myfmt = mdates.DateFormatter("%b %d")
    ax.xaxis.set_major_formatter(myfmt)
    plt.close()
    return fig


def plot_aoe_res_time(
    aoe_class, data, time_param="AoE_Timecorr", figsize=(12, 8), fontsize=12
):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize
    fig, ax = plt.subplots(1, 1)
    try:
        ax.errorbar(
            [
                datetime.strptime(tstamp, "%Y%m%dT%H%M%SZ")
                for tstamp in aoe_class.timecorr_df.index
            ],
            aoe_class.timecorr_df["res"],
            yerr=aoe_class.timecorr_df["res_err"],
            linestyle=" ",
            marker="x",
        )
    except Exception:
        pass
    ax.set_xlabel("time")
    ax.set_ylabel("A/E res")
    myfmt = mdates.DateFormatter("%b %d")
    ax.xaxis.set_major_formatter(myfmt)
    plt.close()
    return fig


def drifttime_corr_plot(
    aoe_class,
    data,
    aoe_param="AoE_Timecorr",
    aoe_param_corr="AoE_DTcorr",
    figsize=(12, 8),
    fontsize=12,
):
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig = plt.figure()

    try:
        dep_events = data.query(
            f"{aoe_class.fit_selection}&{aoe_class.cal_energy_param}>1582&{aoe_class.cal_energy_param}<1602&{aoe_class.cal_energy_param}=={aoe_class.cal_energy_param}&{aoe_param}=={aoe_param}"
        )
        final_df = dep_events.query(aoe_class.dt_res_dict["final_selection"])

        plt.subplot(2, 2, 1)
        aoe_pars = aoe_class.dt_res_dict["aoe_fit1"]["pars"]

        xs = np.linspace(aoe_pars["x_lo"], aoe_pars["x_hi"], 100)
        counts, aoe_bins, bars = plt.hist(
            final_df.query(
                f'{aoe_class.dt_res_dict["aoe_grp1"]}&({aoe_param}<{aoe_pars["x_hi"]})&({aoe_param}>{aoe_pars["x_lo"]})'
            )[aoe_param],
            bins=400,
            histtype="step",
            label="data",
        )
        dx = np.diff(aoe_bins)
        aoe_class.pdf.components = False
        plt.plot(
            xs, aoe_class.pdf.get_pdf(xs, *aoe_pars.values()) * dx[0], label="full fit"
        )
        aoe_class.pdf.components = True
        sig, bkg = aoe_class.pdf.get_pdf(xs, *aoe_pars.values())
        aoe_class.pdf.components = False
        plt.plot(xs, sig * dx[0], label="peak fit")
        plt.plot(xs, bkg * dx[0], label="bkg fit")
        plt.legend(loc="upper left")
        plt.xlabel("A/E")
        plt.ylabel("counts")

        aoe_pars2 = aoe_class.dt_res_dict["aoe_fit2"]["pars"]
        plt.subplot(2, 2, 2)
        xs = np.linspace(aoe_pars2["x_lo"], aoe_pars2["x_hi"], 100)
        counts, aoe_bins2, bars = plt.hist(
            final_df.query(
                f'{aoe_class.dt_res_dict["aoe_grp2"]}&({aoe_param}<{aoe_pars2["x_hi"]})&({aoe_param}>{aoe_pars2["x_lo"]})'
            )[aoe_param],
            bins=400,
            histtype="step",
            label="Data",
        )
        dx = np.diff(aoe_bins2)
        plt.plot(
            xs, aoe_class.pdf.get_pdf(xs, *aoe_pars2.values()) * dx[0], label="full fit"
        )
        aoe_class.pdf.components = True
        sig, bkg = aoe_class.pdf.get_pdf(xs, *aoe_pars2.values())
        aoe_class.pdf.components = False
        plt.plot(xs, sig * dx[0], label="peak fit")
        plt.plot(xs, bkg * dx[0], label="bkg fit")
        plt.legend(loc="upper left")
        plt.xlabel("A/E")
        plt.ylabel("counts")

        hist, bins, var = pgh.get_hist(
            final_df[aoe_class.dt_param],
            dx=32,
            range=(
                np.nanmin(final_df[aoe_class.dt_param]),
                np.nanmax(final_df[aoe_class.dt_param]),
            ),
        )

        plt.subplot(2, 2, 3)
        plt.step(pgh.get_bin_centers(bins), hist, label="data")

        mus = aoe_class.dt_res_dict["dt_fit"]["mus"]
        sigmas = aoe_class.dt_res_dict["dt_fit"]["sigmas"]
        amps = aoe_class.dt_res_dict["dt_fit"]["amps"]

        for mu, sigma, amp in zip(mus, sigmas, amps):
            plt.plot(
                pgh.get_bin_centers(bins),
                nb_gauss_amp(pgh.get_bin_centers(bins), mu, sigma, amp),
            )
        plt.xlabel("drift time (ns)")
        plt.ylabel("Counts")

        plt.subplot(2, 2, 4)
        bins = np.linspace(
            np.nanpercentile(final_df[aoe_param], 1),
            np.nanpercentile(final_df[aoe_param_corr], 99),
            200,
        )
        plt.hist(final_df[aoe_param], bins=bins, histtype="step", label="uncorrected")
        plt.hist(
            final_df[aoe_param_corr], bins=bins, histtype="step", label="corrected"
        )
        plt.xlabel("A/E")
        plt.ylabel("counts")
        plt.legend(loc="upper left")
        plt.tight_layout()
        plt.xlim(bins[0], bins[-1])
    except Exception:
        pass
    plt.close()
    return fig


def plot_compt_bands_overlayed(
    aoe_class,
    data,
    eranges: list[tuple],
    aoe_param="AoE_Timecorr",
    aoe_range: list[float] = None,
    title="Compton Bands",
    density=True,
    n_bins=50,
    figsize=(12, 8),
    fontsize=12,
) -> None:
    """
    Function to plot various compton bands to check energy dependence and corrections
    """
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig = plt.figure()

    for erange in eranges:
        try:
            select_df = data.query(
                f"{aoe_class.selection_string}&{aoe_class.cal_energy_param}>{erange[0]}&{aoe_class.cal_energy_param}<{erange[1]}&{aoe_param}=={aoe_param}"
            )
            if aoe_range is not None:
                select_df = select_df.query(
                    f"{aoe_param}>{aoe_range[0]}&{aoe_param}<{aoe_range[1]}"
                )
                bins = np.linspace(aoe_range[0], aoe_range[1], n_bins)
            else:
                bins = np.linspace(0.85, 1.05, n_bins)
            plt.hist(
                select_df[aoe_param],
                bins=bins,
                histtype="step",
                label=f"{erange[0]}-{erange[1]}",
                density=density,
            )
        except Exception:
            pass
    plt.ylabel("counts")
    plt.xlabel(aoe_param)
    plt.title(title)
    plt.legend(loc="upper left")
    plt.close()
    return fig


def plot_dt_dep(
    aoe_class,
    data,
    eranges: list[tuple],
    titles: list = None,
    aoe_param="AoE_Timecorr",
    bins=(200, 100),
    dt_max=2000,
    figsize=(12, 8),
    fontsize=12,
) -> None:
    """
    Function to produce 2d histograms of A/E against drift time to check dependencies
    """
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig = plt.figure()
    for i, erange in enumerate(eranges):
        try:
            plt.subplot(3, 2, i + 1)
            select_df = data.query(
                f"{aoe_class.selection_string}&{aoe_class.cal_energy_param}<{erange[1]}&{aoe_class.cal_energy_param}>{erange[0]}&{aoe_param}=={aoe_param}"
            )

            hist, bs, var = pgh.get_hist(select_df[aoe_param], bins=500)
            bin_cs = (bs[1:] + bs[:-1]) / 2
            mu = bin_cs[np.argmax(hist)]
            aoe_range = [mu * 0.9, mu * 1.1]

            final_df = select_df.query(
                f"{aoe_param}<{aoe_range[1]}&{aoe_param}>{aoe_range[0]}&{aoe_class.dt_param}<{dt_max}"
            )
            plt.hist2d(
                final_df[aoe_param],
                final_df[aoe_class.dt_param],
                bins=bins,
                norm=LogNorm(),
            )
            plt.ylabel("drift time (ns)")
            plt.xlabel("A/E")
            if titles is None:
                plt.title(f"{erange[0]}-{erange[1]}")
            else:
                plt.title(titles[i])
        except Exception:
            pass
    plt.tight_layout()
    plt.close()
    return fig


def plot_mean_fit(aoe_class, data, figsize=(12, 8), fontsize=12) -> plt.figure:
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize
    fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True)
    try:
        ax1.errorbar(
            aoe_class.energy_corr_fits.index,
            aoe_class.energy_corr_fits["mean"],
            yerr=aoe_class.energy_corr_fits["mean_err"],
            xerr=aoe_class.compt_bands_width / 2,
            label="data",
            linestyle=" ",
        )

        ax1.plot(
            aoe_class.energy_corr_fits.index.to_numpy(),
            aoe_class.mean_func.func(
                aoe_class.energy_corr_fits.index.to_numpy(),
                **aoe_class.energy_corr_res_dict["mean_fits"]["pars"],
            ),
            label="linear model",
        )
        ax1.errorbar(
            1592,
            aoe_class.energy_corr_res_dict["dep_fit"]["pars"]["mu"],
            yerr=aoe_class.energy_corr_res_dict["dep_fit"]["errs"]["mu"],
            label="DEP",
            color="green",
            linestyle=" ",
        )

        ax1.legend(title="A/E mu energy dependence", frameon=False)

        ax1.set_ylabel("raw A/E (a.u.)", ha="right", y=1)
        ax2.scatter(
            aoe_class.energy_corr_fits.index,
            100
            * (
                aoe_class.energy_corr_fits["mean"]
                - aoe_class.mean_func.func(
                    aoe_class.energy_corr_fits.index,
                    **aoe_class.energy_corr_res_dict["mean_fits"]["pars"],
                )
            )
            / aoe_class.mean_func.func(
                aoe_class.energy_corr_fits.index,
                **aoe_class.energy_corr_res_dict["mean_fits"]["pars"],
            ),
            lw=1,
            c="b",
        )
        ax2.scatter(
            1592,
            100
            * (
                aoe_class.energy_corr_res_dict["dep_fit"]["pars"]["mu"]
                - aoe_class.mean_func.func(
                    1592, **aoe_class.energy_corr_res_dict["mean_fits"]["pars"]
                )
            )
            / aoe_class.mean_func.func(
                1592, **aoe_class.energy_corr_res_dict["mean_fits"]["pars"]
            ),
            lw=1,
            c="g",
        )
    except Exception:
        pass
    ax2.set_ylabel("residuals %", ha="right", y=1)
    ax2.set_xlabel("energy (keV)", ha="right", x=1)
    plt.tight_layout()
    plt.close()
    return fig


def plot_sigma_fit(aoe_class, data, figsize=(12, 8), fontsize=12) -> plt.figure:
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True)
    try:
        ax1.errorbar(
            aoe_class.energy_corr_fits.index,
            aoe_class.energy_corr_fits["sigma"],
            yerr=aoe_class.energy_corr_fits["sigma_err"],
            xerr=aoe_class.compt_bands_width / 2,
            label="data",
            linestyle=" ",
        )
        sig_pars = aoe_class.energy_corr_res_dict["SigmaFits"]["pars"]
        if aoe_class.sigma_func == SigmaFit:
            label = f'sqrt model: \nsqrt({sig_pars["a"]:1.4f}+({sig_pars["b"]:1.1f}/E)^{sig_pars["c"]:1.1f})'
        else:
            raise ValueError("unknown sigma function")
        ax1.plot(
            aoe_class.energy_corr_fits.index.to_numpy(),
            aoe_class.sigma_func.func(
                aoe_class.energy_corr_fits.index.to_numpy(), **sig_pars
            ),
            label=label,
        )
        ax1.errorbar(
            1592,
            aoe_class.energy_corr_res_dict["dep_fit"]["pars"]["sigma"],
            yerr=aoe_class.energy_corr_res_dict["dep_fit"]["errs"]["sigma"],
            label="DEP",
            color="green",
            linestyle=" ",
        )
        ax1.set_ylabel("A/E stdev (a.u.)", ha="right", y=1)
        ax1.legend(title="A/E stdev energy dependence", frameon=False)
        ax2.scatter(
            aoe_class.energy_corr_fits.index,
            100
            * (
                aoe_class.energy_corr_fits["sigma"]
                - aoe_class.sigma_func.func(
                    aoe_class.energy_corr_fits.index, **sig_pars
                )
            )
            / aoe_class.sigma_func.func(aoe_class.energy_corr_fits.index, **sig_pars),
            lw=1,
            c="b",
        )
        ax2.scatter(
            1592,
            100
            * (
                aoe_class.energy_corr_res_dict["dep_fit"]["pars"]["sigma"]
                - aoe_class.sigma_func.func(1592, **sig_pars)
            )
            / aoe_class.sigma_func.func(1592, **sig_pars),
            lw=1,
            c="g",
        )
    except Exception:
        pass
    ax2.set_ylabel("residuals", ha="right", y=1)
    ax2.set_xlabel("energy (keV)", ha="right", x=1)
    plt.tight_layout()
    plt.close()
    return fig


def plot_cut_fit(
    aoe_class, data, dep_acc=0.9, figsize=(12, 8), fontsize=12
) -> plt.figure:
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize
    fig = plt.figure()
    try:
        plt.errorbar(
            aoe_class.cut_fits.index,
            aoe_class.cut_fits["sf"],
            yerr=aoe_class.cut_fits["sf_err"],
            linestyle=" ",
        )

        plt.plot(
            aoe_class.cut_fits.index.to_numpy(),
            SigmoidFit.func(
                aoe_class.cut_fits.index.to_numpy(), **aoe_class.cut_fit["pars"]
            ),
        )
        plt.hlines(
            (100 * dep_acc),
            -8.1,
            aoe_class.low_cut_val,
            color="red",
            linestyle="--",
        )
        plt.vlines(
            aoe_class.low_cut_val,
            np.nanmin(aoe_class.cut_fits["sf"]) * 0.9,
            (100 * dep_acc),
            color="red",
            linestyle="--",
        )
        plt.xlim([-8.1, 0.1])
        vals, labels = plt.yticks()
        plt.yticks(vals, [f"{x:,.0f} %" for x in vals])
        plt.ylim([np.nanmin(aoe_class.cut_fits["sf"]) * 0.9, 102])
    except Exception:
        pass
    plt.xlabel("cut value")
    plt.ylabel("survival percentage")
    plt.close()
    return fig


def get_peak_label(peak: float) -> str:
    if peak == 2039:
        return "CC @"
    elif peak == 1592.5:
        return "Tl DEP @"
    elif peak == 1620.5:
        return "Bi FEP @"
    elif peak == 2103.53:
        return "Tl SEP @"
    elif peak == 2614.5:
        return "Tl FEP @"


def plot_survival_fraction_curves(
    aoe_class, data, figsize=(12, 8), fontsize=12
) -> plt.figure:
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig = plt.figure()
    try:
        plt.vlines(
            aoe_class.low_cut_val,
            0,
            100,
            label=f"cut value: {aoe_class.low_cut_val:1.2f}",
            color="black",
        )

        for peak, survival_df in aoe_class.low_side_peak_dfs.items():
            try:
                plt.errorbar(
                    survival_df.index,
                    survival_df["sf"],
                    yerr=survival_df["sf_err"],
                    label=f'{get_peak_label(peak)} {peak} keV: {aoe_class.low_side_sfs.loc[peak]["sf"]:2.1f} +/- {aoe_class.low_side_sfs.loc[peak]["sf_err"]:2.1f} %',
                )
            except Exception:
                pass
    except Exception:
        pass
    vals, labels = plt.yticks()
    plt.yticks(vals, [f"{x:,.0f} %" for x in vals])
    plt.legend(loc="upper right")
    plt.xlabel("cut value")
    plt.ylabel("survival percentage")
    plt.ylim([0, 105])
    plt.close()
    return fig


def plot_spectra(
    aoe_class,
    data,
    xrange=(900, 3000),
    n_bins=2101,
    xrange_inset=(1580, 1640),
    n_bins_inset=200,
    figsize=(12, 8),
    fontsize=12,
) -> plt.figure:
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig, ax = plt.subplots()
    try:
        bins = np.linspace(xrange[0], xrange[1], n_bins)
        ax.hist(
            data.query(aoe_class.selection_string)[aoe_class.cal_energy_param],
            bins=bins,
            histtype="step",
            label="before PSD",
        )
        ax.hist(
            data.query(f"{aoe_class.selection_string}&AoE_Low_Cut")[
                aoe_class.cal_energy_param
            ],
            bins=bins,
            histtype="step",
            label="low side PSD cut",
        )
        ax.hist(
            data.query(f"{aoe_class.selection_string}&AoE_Double_Sided_Cut")[
                aoe_class.cal_energy_param
            ],
            bins=bins,
            histtype="step",
            label="double sided PSD cut",
        )
        ax.hist(
            data.query(f"{aoe_class.selection_string} & (~AoE_Double_Sided_Cut)")[
                aoe_class.cal_energy_param
            ],
            bins=bins,
            histtype="step",
            label="rejected by PSD cut",
        )

        axins = ax.inset_axes([0.25, 0.07, 0.4, 0.3])
        bins = np.linspace(xrange_inset[0], xrange_inset[1], n_bins_inset)
        select_df = data.query(
            f"{aoe_class.cal_energy_param}<{xrange_inset[1]}&{aoe_class.cal_energy_param}>{xrange_inset[0]}"
        )
        axins.hist(
            select_df.query(aoe_class.selection_string)[aoe_class.cal_energy_param],
            bins=bins,
            histtype="step",
        )
        axins.hist(
            select_df.query(f"{aoe_class.selection_string}&AoE_Low_Cut")[
                aoe_class.cal_energy_param
            ],
            bins=bins,
            histtype="step",
        )
        axins.hist(
            select_df.query(f"{aoe_class.selection_string}&AoE_Double_Sided_Cut")[
                aoe_class.cal_energy_param
            ],
            bins=bins,
            histtype="step",
        )
        axins.hist(
            select_df.query(f"{aoe_class.selection_string} & (~AoE_Double_Sided_Cut)")[
                aoe_class.cal_energy_param
            ],
            bins=bins,
            histtype="step",
        )
    except Exception:
        pass
    ax.set_xlim(xrange)
    ax.set_yscale("log")
    plt.xlabel("energy (keV)")
    plt.ylabel("counts")
    plt.legend(loc="upper left")
    plt.close()
    return fig


def plot_sf_vs_energy(
    aoe_class, data, xrange=(900, 3000), n_bins=701, figsize=(12, 8), fontsize=12
) -> plt.figure:
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig = plt.figure()
    try:
        bins = np.linspace(xrange[0], xrange[1], n_bins)
        counts_pass, bins_pass, _ = pgh.get_hist(
            data.query(f"{aoe_class.selection_string}&AoE_Double_Sided_Cut")[
                aoe_class.cal_energy_param
            ],
            bins=bins,
        )
        counts, bins, _ = pgh.get_hist(
            data.query(aoe_class.selection_string)[aoe_class.cal_energy_param],
            bins=bins,
        )
        survival_fracs = counts_pass / (counts + 10**-99)

        plt.step(pgh.get_bin_centers(bins_pass), 100 * survival_fracs)
    except Exception:
        pass
    plt.ylim([0, 100])
    vals, labels = plt.yticks()
    plt.yticks(vals, [f"{x:,.0f} %" for x in vals])
    plt.xlabel("energy (keV)")
    plt.ylabel("survival percentage")
    plt.close()
    return fig


def plot_classifier(
    aoe_class,
    data,
    aoe_param="AoE_Classifier",
    xrange=(900, 3000),
    yrange=(-50, 10),
    xn_bins=700,
    yn_bins=500,
    figsize=(12, 8),
    fontsize=12,
) -> plt.figure:
    plt.rcParams["figure.figsize"] = figsize
    plt.rcParams["font.size"] = fontsize

    fig = plt.figure()
    try:
        plt.hist2d(
            data.query(aoe_class.selection_string)[aoe_class.cal_energy_param],
            data.query(aoe_class.selection_string)[aoe_param],
            bins=[
                np.linspace(xrange[0], xrange[1], xn_bins),
                np.linspace(yrange[0], yrange[1], yn_bins),
            ],
            norm=LogNorm(),
        )
    except Exception:
        pass
    plt.xlabel("energy (keV)")
    plt.ylabel(aoe_param)
    plt.xlim(xrange)
    plt.ylim(yrange)
    plt.close()
    return fig



import numpy as np
from iminuit import Minuit, cost
from pytest import approx
from scipy.stats import exponnorm

from pygama.math.distributions import exgauss, hpge_peak

xr = (-20, 20)  # xrange

tau = 1.1
sigma = 1
mu = 2
area = 1000
scipy_exgauss = exponnorm(tau / sigma, -mu, sigma)

xdata = scipy_exgauss.rvs(size=1000, random_state=1)
xmix = -1 * xdata

n, xe = np.histogram(xmix, bins=50, range=xr)
cx = 0.5 * (xe[1:] + xe[:-1])
dx = np.diff(xe)


def test_unbinned_nll():
    c = cost.UnbinnedNLL(xmix, exgauss.pdf_norm)

    m = Minuit(c, x_lo=-20, x_hi=20, tau=1, mu=0, sigma=3)
    m.fixed["x_lo", "x_hi"] = True
    m.limits["tau"] = (0.1, 3)
    m.limits["mu"] = (0, 2)
    m.limits["sigma"] = (0, 1.5)
    m.migrad()

    fit_tau = m.values["tau"]
    fit_mu = m.values["mu"]
    fit_sigma = m.values["sigma"]

    assert tau == approx(fit_tau, 0.1)
    assert mu == approx(fit_mu, 0.1)
    assert sigma == approx(fit_sigma, 0.1)


def test_extended_unbinned_nll():
    c = cost.ExtendedUnbinnedNLL(xmix, exgauss.pdf_ext)

    m = Minuit(c, x_lo=-7, x_hi=7, area=100, mu=0, sigma=3, tau=1)
    m.fixed["x_lo", "x_hi"] = True
    m.limits["tau", "mu", "sigma"] = (0.01, 5)
    m.limits["area"] = (0, 2000)
    m.migrad()

    fit_area = m.values["area"]
    fit_tau = m.values["tau"]
    fit_mu = m.values["mu"]
    fit_sigma = m.values["sigma"]

    assert area == approx(fit_area, 0.1)
    assert tau == approx(fit_tau, 0.1)
    assert mu == approx(fit_mu, 0.1)
    assert sigma == approx(fit_sigma, 0.1)


def test_binned_nll():
    c = cost.BinnedNLL(n, xe, exgauss.cdf_norm)

    m = Minuit(c, x_lo=-20, x_hi=20, mu=0, sigma=3, tau=1)
    m.fixed["x_lo", "x_hi"] = True
    m.limits["tau"] = (0.1, 3)
    m.limits["mu"] = (0, 2)
    m.limits["sigma"] = (0, 1.5)
    m.migrad()

    fit_tau = m.values["tau"]
    fit_mu = m.values["mu"]
    fit_sigma = m.values["sigma"]

    assert tau == approx(fit_tau, 0.1)
    assert mu == approx(fit_mu, 0.1)
    assert sigma == approx(fit_sigma, 0.1)


def test_extended_binned_nll():
    c = cost.ExtendedBinnedNLL(n, xe, exgauss.cdf_ext)

    m = Minuit(c, area=100, mu=0, sigma=3, tau=1)
    m.limits["tau", "mu", "sigma"] = (0.01, 5)
    m.limits["area"] = (0, 2000)

    m.migrad()

    fit_area = m.values["area"]
    fit_tau = m.values["tau"]
    fit_mu = m.values["mu"]
    fit_sigma = m.values["sigma"]

    assert area == approx(fit_area, 0.1)
    assert tau == approx(fit_tau, 0.1)
    assert mu == approx(fit_mu, 0.1)
    assert sigma == approx(fit_sigma, 0.1)


def test_name_introspection():
    c = cost.ExtendedUnbinnedNLL(xmix, hpge_peak.pdf_ext)
    m = Minuit(c, xr[0], xr[-1], 100, 1, 2, 0.1, 20, 10, 0.1)

    m.values["n_sig"] = 0
    m.values["mu"] = 0
    m.values["sigma"] = 0
    m.values["htail"] = 0
    m.values["tau"] = 0
    m.values["n_bkg"] = 0
    m.values["hstep"] = 0
    m.fixed["x_lo"] = 0
    m.fixed["x_hi"] = 0



import logging
import sys
from collections import namedtuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pint
from dspeed import build_processing_chain
from dspeed.units import unit_registry as ureg
from matplotlib.colors import LogNorm
from scipy.optimize import minimize
from scipy.stats import norm
from sklearn.exceptions import ConvergenceWarning
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.utils._testing import ignore_warnings

log = logging.getLogger(__name__)


def run_one_dsp(
    tb_data, dsp_config, db_dict=None, fom_function=None, verbosity=0, fom_kwargs=None
):
    """
    run one iteration of DSP on tb_data

    Optionally returns a value for optimization

    Parameters
    ----------
    tb_data : lh5 Table
        An input table of lh5 data. Typically a selection is made prior to
        sending tb_data to this function: optimization typically doesn't have to
        run over all data
    dsp_config : dict
        Specifies the DSP to be performed for this iteration (see
        build_processing_chain()) and the list of output variables to appear in
        the output table
    db_dict : dict (optional)
        DSP parameters database. See build_processing_chain for formatting info
    fom_function : function or None (optional)
        When given the output lh5 table of this DSP iteration, the
        fom_function must return a scalar figure-of-merit value upon which the
        optimization will be based. Should accept verbosity as a second argument
    verbosity : int (optional)
        verbosity for the processing chain and fom_function calls
    fom_kwargs
        any keyword arguments to pass to the fom

    Returns
    -------
    figure_of_merit : float
        If fom_function is not None, returns figure-of-merit value for the DSP iteration
    tb_out : lh5 Table
        If fom_function is None, returns the output lh5 table for the DSP iteration
    """

    pc, lh5_col_names, tb_out = build_processing_chain(
        tb_data, dsp_config, db_dict=db_dict
    )
    pc.execute()
    if fom_function is not None:
        if fom_kwargs is not None:
            return fom_function(tb_out, verbosity, fom_kwargs)
        else:
            return fom_function(tb_out, verbosity)
    else:
        return tb_out


ParGridDimension = namedtuple("ParGridDimension", "name parameter value_strs")


class ParGrid:
    """Parameter Grid class
    Each ParGrid entry corresponds to a dsp parameter to be varied.
    The ntuples must follow the pattern:
    ( name parameter value_strs) : ( str, str, list of str)
    where name and parameter are the same as 'db.name.parameter' in the processing chain,
    value_strs is the array of strings to set the argument to.
    """

    def __init__(self):
        self.dims = []

    def add_dimension(self, name, parameter, value_strs):
        self.dims.append(ParGridDimension(name, parameter, value_strs))

    def get_n_dimensions(self):
        return len(self.dims)

    def get_n_points_of_dim(self, i):
        return len(self.dims[i].value_strs)

    def get_shape(self):
        shape = ()
        for i in range(self.get_n_dimensions()):
            shape += (self.get_n_points_of_dim(i),)
        return shape

    def get_n_grid_points(self):
        return np.prod(self.get_shape())

    def get_par_meshgrid(self, copy=False, sparse=False):
        """return a meshgrid of parameter values
        Always uses Matrix indexing (natural for par grid) so that
        mg[i1][i2][...] corresponds to index order in self.dims
        Note copy is False by default as opposed to numpy default of True
        """
        axes = []
        for i in range(self.get_n_dimensions()):
            axes.append(self.dims[i].values_strs)
        return np.meshgrid(*axes, copy, sparse, indexing="ij")

    def get_zero_indices(self):
        return np.zeros(self.get_n_dimensions(), dtype=np.uint32)

    def iterate_indices(self, indices):
        """iterate given indices [i1, i2, ...] by one.
        For easier iteration. The convention here is arbitrary, but its the
        order the arrays would be traversed in a series of nested for loops in
        the order appearing in dims (first dimension is first for loop, etc):
        Return False when the grid runs out of indices. Otherwise returns True.
        """
        for dim in reversed(range(self.get_n_dimensions())):
            indices[dim] += 1
            if indices[dim] < self.get_n_points_of_dim(dim):
                return True
            indices[dim] = 0
        return False

    # def check_indices(self, indices):
    #    for iD in reversed(range(self.get_n_dimensions())):
    #        if indices[iD] < self.get_n_points_of_dim(iD): return True
    #        indices[iD] = 0
    #    return False

    def get_data(self, i_dim, i_par):
        name = self.dims[i_dim].name
        parameter = self.dims[i_dim].parameter
        value_str = self.dims[i_dim].value_strs[i_par]
        return name, parameter, value_str

    def print_data(self, indices):
        print_string = f"Grid point at indices {indices}:"
        for i_dim, i_par in enumerate(indices):
            name, parameter, value_str = self.get_data(i_dim, i_par)
            print_string += f"\n {name}.{parameter} = {value_str}"
        return print_string

    def set_dsp_pars(self, db_dict, indices):
        if db_dict is None:
            db_dict = {}
        for i_dim, i_par in enumerate(indices):
            name, parameter, value_str = self.get_data(i_dim, i_par)
            if name not in db_dict.keys():
                db_dict[name] = {parameter: value_str}
            else:
                db_dict[name][parameter] = value_str
        return db_dict


def run_grid(
    tb_data, dsp_config, grid, fom_function, db_dict=None, verbosity=1, **fom_kwargs
):
    """Extract a table of optimization values for a grid of DSP parameters
    The grid argument defines a list of parameters and values over which to run
    the DSP defined in dsp_config on tb_data. At each point, a scalar
    figure-of-merit is extracted.

    Returns a N-dimensional ndarray of figure-of-merit values, where the array
    axes are in the order they appear in grid.

    Parameters
    ----------
    tb_data : lh5 Table
        An input table of lh5 data. Typically a selection is made prior to
        sending tb_data to this function: optimization typically doesn't have to
        run over all data
    dsp_config : dict
        Specifies the DSP to be performed (see build_processing_chain()) and the
        list of output variables to appear in the output table for each grid point
    grid : ParGrid
        See ParGrid class for format
    fom_function : function
        When given the output lh5 table of this DSP iteration, the fom_function
        must return a scalar figure-of-merit. Should accept verbosity as a
        second keyword argument
    db_dict : dict (optional)
        DSP parameters database. See build_processing_chain for formatting info
    verbosity : int (optional)
        verbosity for the processing chain and fom_function calls
    **fom_kwargs
        Any keyword arguments for fom_function

    Returns
    -------
    grid_values : ndarray of floats
        An N-dimensional numpy ndarray whose Mth axis corresponds to the Mth row
        of the grid argument
    """

    grid_values = np.ndarray(shape=grid.get_shape(), dtype="O")
    iii = grid.get_zero_indices()
    log.info("starting grid calculations...")
    while True:
        db_dict = grid.set_dsp_pars(db_dict, iii)
        if verbosity > 1:
            log.debug(dsp_config)
        log.debug(grid.print_data(iii))
        grid_values[tuple(iii)] = run_one_dsp(
            tb_data,
            dsp_config,
            db_dict=db_dict,
            fom_function=fom_function,
            verbosity=verbosity,
            fom_kwargs=fom_kwargs,
        )
        log.debug("value:", grid_values[tuple(iii)])
        if not grid.iterate_indices(iii):
            break
    return grid_values


def run_grid_point(
    tb_data,
    dsp_config,
    grids,
    fom_function,
    iii,
    db_dict=None,
    verbosity=1,
    fom_kwargs=None,
):
    """
    Runs a single grid point for the index specified
    """
    if not isinstance(grids, list):
        grids = [grids]
    for index, grid in zip(iii, grids):
        db_dict = grid.set_dsp_pars(db_dict, index)

    log.debug(db_dict)
    for i, grid in enumerate(grids):
        log.debug(grid.print_data(iii[i]))
    tb_out = run_one_dsp(tb_data, dsp_config, db_dict=db_dict, verbosity=verbosity)
    res = np.ndarray(shape=len(grids), dtype="O")
    if fom_function:
        for i in range(len(grids)):
            if fom_kwargs[i] is not None:
                if len(fom_function) > 1:
                    res[i] = fom_function[i](tb_out, verbosity, fom_kwargs[i])
                else:
                    res[i] = fom_function[0](tb_out, verbosity, fom_kwargs[i])
            else:
                if len(fom_function) > 1:
                    res[i] = fom_function[i](tb_out, verbosity)
                else:
                    res[i] = fom_function[0](tb_out, verbosity)
        log.debug("value:", res)
        out = {"indexes": [tuple(ii) for ii in iii], "results": res}

    else:
        out = {"indexes": [tuple(ii) for ii in iii], "results": tb_out}
    return out


def get_grid_points(grid):
    """
    Generates a list of the indices of all possible grid points
    """
    out = []
    iii = [gri.get_zero_indices() for gri in grid]
    complete = np.full(len(grid), False)
    while True:
        out.append([tuple(ii) for ii in iii])

        for i, gri in enumerate(zip(iii, grid)):
            if not gri[1].iterate_indices(gri[0]):
                log.info(f"{i} grid end")
                iii[i] = gri[1].get_zero_indices()
                complete[i] = True
        if all(complete):
            break
    return out


OptimiserDimension = namedtuple(
    "OptimiserDimension", "name parameter min_val max_val round unit"
)


class BayesianOptimizer:
    """
    Bayesian optimiser uses Gaussian Process Regressor from sklearn to fit kernel
    to data, takes in a series of init samples for this fit and then calculates
    the next point using the acquisition function specified.
    """

    np.random.seed(55)
    lambda_param = 0.01
    eta_param = 0

    def __init__(
        self,
        acq_func,
        batch_size,
        kernel=None,
        sampling_rate=None,
        fom_value="y_val",
        fom_error="y_val_err",
    ):
        self.dims = []
        self.current_iter = 0

        self.batch_size = batch_size
        self.iters = 0

        if isinstance(sampling_rate, str):
            self.sampling_rate = ureg.Quantity(sampling_rate)
        elif isinstance(sampling_rate, pint.Quantity):
            self.sampling_rate = sampling_rate
        else:
            if sampling_rate is not None:
                raise TypeError("Unknown type for sampling rate")

        self.gauss_pr = GaussianProcessRegressor(kernel=kernel)
        self.best_samples_ = pd.DataFrame(columns=["x", "y", "ei"])
        self.distances_ = []

        if acq_func == "ei":
            self.acq_function = self._get_expected_improvement
        elif acq_func == "ucb":
            self.acq_function = self._get_ucb
        elif acq_func == "lcb":
            self.acq_function = self._get_lcb

        self.fom_value = fom_value
        self.fom_error = fom_error

    def add_dimension(
        self, name, parameter, min_val, max_val, round_to_samples=False, unit=None
    ):
        if round_to_samples is True and self.sampling_rate is None:
            raise ValueError("Must provide sampling rate to round to samples")
        if unit is not None:
            unit = ureg.Quantity(unit)
        self.dims.append(
            OptimiserDimension(
                name, parameter, min_val, max_val, round_to_samples, unit
            )
        )

    def get_n_dimensions(self):
        return len(self.dims)

    def add_initial_values(self, x_init, y_init, yerr_init):
        self.x_init = x_init
        self.y_init = y_init
        self.yerr_init = yerr_init

    def _get_expected_improvement(self, x_new):
        mean_y_new, sigma_y_new = self.gauss_pr.predict(
            np.array([x_new]), return_std=True
        )

        mean_y = self.gauss_pr.predict(self.x_init)
        min_mean_y = np.min(mean_y)
        z = (mean_y_new[0] - min_mean_y - 1) / (sigma_y_new[0] + 1e-9)
        exp_imp = (mean_y_new[0] - min_mean_y - 1) * norm.cdf(z) + sigma_y_new[
            0
        ] * norm.pdf(z)
        return exp_imp

    def _get_ucb(self, x_new):
        mean_y_new, sigma_y_new = self.gauss_pr.predict(
            np.array([x_new]), return_std=True
        )
        return mean_y_new[0] + self.lambda_param * sigma_y_new[0]

    def _get_lcb(self, x_new):
        mean_y_new, sigma_y_new = self.gauss_pr.predict(
            np.array([x_new]), return_std=True
        )
        return mean_y_new[0] - self.lambda_param * sigma_y_new[0]

    def _get_next_probable_point(self):
        min_ei = float(sys.maxsize)
        x_optimal = None
        # Trial with an array of random data points
        rands = np.random.uniform(
            np.array([dim.min_val for dim in self.dims]),
            np.array([dim.max_val for dim in self.dims]),
            (self.batch_size, self.get_n_dimensions()),
        )
        for x_start in rands:
            response = minimize(
                fun=self.acq_function,
                x0=x_start,
                bounds=[(dim.min_val, dim.max_val) for dim in self.dims],
                method="L-BFGS-B",
            )
            if response.fun < min_ei:
                min_ei = response.fun
                x_optimal = []
                for y, dim in zip(response.x, self.dims):
                    if dim.round is True and dim.unit is not None:
                        # round so samples is integer

                        x_optimal.append(
                            float(
                                round(
                                    (y * (dim.unit / self.sampling_rate)).to(
                                        "dimensionless"
                                    ),
                                    0,
                                )
                                * (self.sampling_rate / dim.unit)
                            )
                        )
                    else:
                        x_optimal.append(y)
        if x_optimal in self.x_init:
            perturb = np.random.uniform(
                -np.array([(dim.max_val - dim.min_val) / 10 for dim in self.dims]),
                np.array([(dim.max_val - dim.min_val) / 10 for dim in self.dims]),
                (1, len(self.dims)),
            )
            x_optimal += perturb
            new_x_optimal = []
            for y, dim in zip(x_optimal[0], self.dims):
                if dim.round is True and dim.unit is not None:
                    # round so samples is integer
                    new_x_optimal.append(
                        float(
                            round(
                                (y * (dim.unit / self.sampling_rate)).to(
                                    "dimensionless"
                                ),
                                0,
                            )
                            * (self.sampling_rate / dim.unit)
                        )
                    )
                else:
                    new_x_optimal.append(y)
            x_optimal = new_x_optimal
            for i, y in enumerate(x_optimal):
                if y > self.dims[i].max_val:
                    x_optimal[i] = self.dims[i].max_val
                elif y < self.dims[i].min_val:
                    x_optimal[i] = self.dims[i].min_val
        return x_optimal, min_ei

    def _extend_prior_with_posterior_data(self, x, y, yerr):
        self.x_init = np.append(self.x_init, np.array([x]), axis=0)
        self.y_init = np.append(self.y_init, np.array(y), axis=0)
        self.yerr_init = np.append(self.yerr_init, np.array(yerr), axis=0)

    def get_first_point(self):
        y_min_ind = np.nanargmin(self.y_init)
        self.y_min = self.y_init[y_min_ind]
        self.optimal_x = self.x_init[y_min_ind]
        self.optimal_ei = None
        return self.optimal_x, self.optimal_ei

    @ignore_warnings(category=ConvergenceWarning)
    def iterate_values(self):
        nan_idxs = np.isnan(self.y_init)
        self.gauss_pr.fit(self.x_init[~nan_idxs], np.array(self.y_init)[~nan_idxs])
        x_next, ei = self._get_next_probable_point()
        return x_next, ei

    def update_db_dict(self, db_dict):
        if self.current_iter == 0:
            x_new, ei = self.get_first_point()
        x_new, ei = self.iterate_values()
        self.current_x = x_new
        self.current_ei = ei
        for i, val in enumerate(x_new):
            name, parameter, min_val, max_val, rounding, unit = self.dims[i]
            if unit is not None:
                value_str = f"{val}*{unit.units:~}"
                if "µ" in value_str:
                    value_str = value_str.replace("µ", "u")
            else:
                value_str = f"{val}"
            if name not in db_dict.keys():
                db_dict[name] = {parameter: value_str}
            else:
                db_dict[name][parameter] = value_str
        self.current_iter += 1
        return db_dict

    def update(self, results):
        y_val = results[self.fom_value]
        y_err = results[self.fom_error]
        self._extend_prior_with_posterior_data(
            self.current_x, np.array([y_val]), np.array([y_err])
        )

        if np.isnan(y_val) | np.isnan(y_err):
            pass
        else:
            if y_val < self.y_min:
                self.y_min = y_val
                self.optimal_x = self.current_x
                self.optimal_ei = self.current_ei
                self.optimal_results = results

        if self.current_iter == 1:
            self.prev_x = self.current_x
        else:
            self.distances_.append(
                np.linalg.norm(np.array(self.prev_x) - np.array(self.current_x))
            )
            self.prev_x = self.current_x
        new_entry = pd.DataFrame(
            {"x": self.optimal_x, "y": self.y_min, "ei": self.optimal_ei}
        )
        if (
            not new_entry.empty
            and new_entry.notnull().any().any()
            and len(new_entry) >= 1
        ):
            self.best_samples_ = pd.concat(
                [
                    self.best_samples_,
                    new_entry,
                ],
                ignore_index=True,
            )

    def get_best_vals(self):
        out_dict = {}
        for i, val in enumerate(self.optimal_x):
            name, parameter, min_val, max_val, rounding, unit = self.dims[i]
            if unit is not None:
                value_str = f"{val}*{unit.units:~}"
                if "µ" in value_str:
                    value_str = value_str.replace("µ", "u")
            else:
                value_str = f"{val}"
            if name not in out_dict.keys():
                out_dict[name] = {parameter: value_str}
            else:
                out_dict[name][parameter] = value_str
        return out_dict

    @ignore_warnings(category=ConvergenceWarning)
    def plot(self, init_samples=None):
        nan_idxs = np.isnan(self.y_init)
        fail_idxs = np.isnan(self.yerr_init)
        self.gauss_pr.fit(self.x_init[~nan_idxs], np.array(self.y_init)[~nan_idxs])
        if (len(self.dims) != 2) and (len(self.dims) != 1):
            raise Exception("Acquisition Function Plotting not implemented for dim!=2")
        elif len(self.dims) == 1:
            points = np.arange(self.dims[0].min_val, self.dims[0].max_val, 0.1)
            ys = np.zeros_like(points)
            ys_err = np.zeros_like(points)
            for i, point in enumerate(points):
                ys[i], ys_err[i] = self.gauss_pr.predict(
                    np.array([point]).reshape(1, -1), return_std=True
                )
            fig = plt.figure()

            plt.scatter(np.array(self.x_init), np.array(self.y_init), label="Samples")
            plt.scatter(
                np.array(self.x_init)[fail_idxs],
                np.array(self.y_init)[fail_idxs],
                color="green",
                label="Failed samples",
            )
            plt.fill_between(points, ys - ys_err, ys + ys_err, alpha=0.1)
            if init_samples is not None:
                init_ys = np.array(
                    [
                        np.where(init_sample == self.x_init)[0][0]
                        for init_sample in init_samples
                    ]
                )
                plt.scatter(
                    np.array(init_samples)[:, 0],
                    np.array(self.y_init)[init_ys],
                    color="red",
                    label="Init Samples",
                )
            plt.scatter(self.optimal_x[0], self.y_min, color="orange", label="Optimal")

            plt.xlabel(
                f"{self.dims[0].name}-{self.dims[0].parameter}({self.dims[0].unit})"
            )
            plt.ylabel("Kernel Value")
            plt.legend()
        elif len(self.dims) == 2:
            x, y = np.mgrid[
                self.dims[0].min_val : self.dims[0].max_val : 0.1,
                self.dims[1].min_val : self.dims[1].max_val : 0.1,
            ]
            points = np.vstack((x.flatten(), y.flatten())).T
            out_grid = np.zeros(
                (
                    int((self.dims[0].max_val - self.dims[0].min_val) * 10),
                    int((self.dims[1].max_val - self.dims[1].min_val) * 10),
                )
            )

            j = 0
            for i, _ in np.ndenumerate(out_grid):
                out_grid[i] = self.gauss_pr.predict(
                    points[j].reshape(1, -1), return_std=False
                )
                j += 1

            fig = plt.figure()
            plt.imshow(
                out_grid,
                norm=LogNorm(),
                origin="lower",
                aspect="auto",
                extent=(0, out_grid.shape[1], 0, out_grid.shape[0]),
            )
            plt.scatter(
                np.array(self.x_init - self.dims[1].min_val)[:, 1] * 10,
                np.array(self.x_init - self.dims[0].min_val)[:, 0] * 10,
            )
            if init_samples is not None:
                plt.scatter(
                    (init_samples[:, 1] - self.dims[1].min_val) * 10,
                    (init_samples[:, 0] - self.dims[0].min_val) * 10,
                    color="red",
                )
            plt.scatter(
                (self.optimal_x[1] - self.dims[1].min_val) * 10,
                (self.optimal_x[0] - self.dims[0].min_val) * 10,
                color="orange",
            )
            ticks, labels = plt.xticks()
            labels = np.linspace(self.dims[1].min_val, self.dims[1].max_val, 5)
            ticks = np.linspace(0, out_grid.shape[1], 5)
            plt.xticks(ticks=ticks, labels=labels, rotation=45)
            ticks, labels = plt.yticks()
            labels = np.linspace(self.dims[0].min_val, self.dims[0].max_val, 5)
            ticks = np.linspace(0, out_grid.shape[0], 5)
            plt.yticks(ticks=ticks, labels=labels, rotation=45)
            plt.xlabel(
                f"{self.dims[1].name}-{self.dims[1].parameter}({self.dims[1].unit})"
            )
            plt.ylabel(
                f"{self.dims[0].name}-{self.dims[0].parameter}({self.dims[0].unit})"
            )
        plt.title(f"{self.dims[0].name} Kernel Prediction")
        plt.tight_layout()
        plt.close()
        return fig

    @ignore_warnings(category=ConvergenceWarning)
    def plot_acq(self, init_samples=None):
        nan_idxs = np.isnan(self.y_init)
        self.gauss_pr.fit(self.x_init[~nan_idxs], np.array(self.y_init)[~nan_idxs])
        if (len(self.dims) != 2) and (len(self.dims) != 1):
            raise Exception("Acquisition Function Plotting not implemented for dim!=2")
        elif len(self.dims) == 1:
            points = np.arange(self.dims[0].min_val, self.dims[0].max_val, 0.1)
            ys = np.zeros_like(points)
            for i, point in enumerate(points):
                ys[i] = self.acq_function(np.array([point]).reshape(1, -1)[0])
            fig = plt.figure()
            plt.plot(points, ys)
            plt.scatter(np.array(self.x_init), np.array(self.y_init), label="Samples")
            if init_samples is not None:
                init_ys = np.array(
                    [
                        np.where(init_sample == self.x_init)[0][0]
                        for init_sample in init_samples
                    ]
                )
                plt.scatter(
                    np.array(init_samples)[:, 0],
                    np.array(self.y_init)[init_ys],
                    color="red",
                    label="Init Samples",
                )
            plt.scatter(self.optimal_x[0], self.y_min, color="orange", label="Optimal")

            plt.xlabel(
                f"{self.dims[0].name}-{self.dims[0].parameter}({self.dims[0].unit})"
            )
            plt.ylabel("Acquisition Function Value")
            plt.legend()

        elif len(self.dims) == 2:
            x, y = np.mgrid[
                self.dims[0].min_val : self.dims[0].max_val : 0.1,
                self.dims[1].min_val : self.dims[1].max_val : 0.1,
            ]
            points = np.vstack((x.flatten(), y.flatten())).T
            out_grid = np.zeros(
                (
                    int((self.dims[0].max_val - self.dims[0].min_val) * 10),
                    int((self.dims[1].max_val - self.dims[1].min_val) * 10),
                )
            )

            j = 0
            for i, _ in np.ndenumerate(out_grid):
                out_grid[i] = self.acq_function(points[j])
                j += 1

            fig = plt.figure()
            plt.imshow(
                out_grid,
                norm=LogNorm(),
                origin="lower",
                aspect="auto",
                extent=(0, out_grid.shape[1], 0, out_grid.shape[0]),
            )
            plt.scatter(
                np.array(self.x_init - self.dims[1].min_val)[:, 1] * 10,
                np.array(self.x_init - self.dims[0].min_val)[:, 0] * 10,
            )
            if init_samples is not None:
                plt.scatter(
                    (init_samples[:, 1] - self.dims[1].min_val) * 10,
                    (init_samples[:, 0] - self.dims[0].min_val) * 10,
                    color="red",
                )
            plt.scatter(
                (self.optimal_x[1] - self.dims[1].min_val) * 10,
                (self.optimal_x[0] - self.dims[0].min_val) * 10,
                color="orange",
            )
            ticks, labels = plt.xticks()
            labels = np.linspace(self.dims[1].min_val, self.dims[1].max_val, 5)
            ticks = np.linspace(0, out_grid.shape[1], 5)
            plt.xticks(ticks=ticks, labels=labels, rotation=45)
            ticks, labels = plt.yticks()
            labels = np.linspace(self.dims[0].min_val, self.dims[0].max_val, 5)
            ticks = np.linspace(0, out_grid.shape[0], 5)
            plt.yticks(ticks=ticks, labels=labels, rotation=45)
            plt.xlabel(
                f"{self.dims[1].name}-{self.dims[1].parameter}({self.dims[1].unit})"
            )
            plt.ylabel(
                f"{self.dims[0].name}-{self.dims[0].parameter}({self.dims[0].unit})"
            )
        plt.title(f"{self.dims[0].name} Acquisition Space")
        plt.tight_layout()
        plt.close()
        return fig


def run_bayesian_optimisation(
    tb_data,
    dsp_config,
    fom_function,
    optimisers,
    fom_kwargs=None,
    db_dict=None,
    nan_val=10,
    n_iter=10,
):
    if not isinstance(optimisers, list):
        optimisers = [optimisers]
    if not isinstance(fom_kwargs, list):
        fom_kwargs = [fom_kwargs]
    if not isinstance(fom_function, list):
        fom_function = [fom_function]

    for j in range(n_iter):
        for optimiser in optimisers:
            db_dict = optimiser.update_db_dict(db_dict)

        log.info(f"Iteration number: {j+1}")
        log.info(f"Processing with {db_dict}")

        tb_out = run_one_dsp(tb_data, dsp_config, db_dict=db_dict)

        res = np.ndarray(shape=len(optimisers), dtype="O")

        for i in range(len(optimisers)):
            if fom_kwargs[i] is not None:
                if len(fom_function) > 1:
                    res[i] = fom_function[i](tb_out, fom_kwargs[i])
                else:
                    res[i] = fom_function[0](tb_out, fom_kwargs[i])
            else:
                if len(fom_function) > 1:
                    res[i] = fom_function[i](tb_out)
                else:
                    res[i] = fom_function[0](tb_out)

        log.info(f"Results of iteration {j+1} are {res}")

        for i, optimiser in enumerate(optimisers):
            if np.isnan(res[i][optimiser.fom_value]):
                if isinstance(nan_val, list):
                    res[i][optimiser.fom_value] = nan_val[i]
                else:
                    res[i][optimiser.fom_value] = nan_val

            optimiser.update(res[i])

    out_param_dict = {}
    out_results_list = []
    for optimiser in optimisers:
        param_dict = optimiser.get_best_vals()
        out_param_dict.update(param_dict)
        results_dict = optimiser.optimal_results
        if np.isnan(results_dict[optimiser.fom_value]):
            log.error(f"Energy optimisation failed for {optimiser.dims[0][0]}")
        out_results_list.append(results_dict)

    return out_param_dict, out_results_list



import numpy as np
from pytest import approx

import pygama.math.hpge_peak_fitting as pgb


def test_mostly_gauss_fwhm():
    # parameters need to be in order of n_sig, mu, sigma, htail, tau, n_bkg, hstep
    pars = [1, 0, 1, 0, 0.1, 0, 0]
    cov = [
        [1e-16, 0, 0, 0, 0, 0, 0],  # damp2
        [0, 1e-16, 0, 0, 0, 0, 0],  # dmu2
        [0, 0, 1e-02, 0, 0, 0, 0],  # dsig2
        [0, 0, 0, 1e-16, 0, 0, 0],  # dhtail2
        [0, 0, 0, 0, 1e-16, 0, 0],  # dtau2
        [0, 0, 0, 0, 0, 1e-16, 0],  # dbg02
        [0, 0, 0, 0, 0, 0, 1e-16],  # dhstep2
    ]
    amp, mu, sig, htail, tau, bg0, hstep = pars
    fwhm, dfwhm = pgb.hpge_peak_fwhm(sig, htail, tau, cov)
    assert fwhm == approx(2.3548, rel=1e-5)
    assert dfwhm == approx(2.3548e-1, rel=1e-5)


def test_mostly_exp_fwhm():
    # parameters need to be in order of n_sig, mu, sigma, htail, tau, n_bkg, hstep
    pars = [1, 0, 1e-6, 1, 1, 0, 0]
    cov = [
        [1e-16, 0, 0, 0, 0, 0, 0],  # damp2
        [0, 1e-16, 0, 0, 0, 0, 0],  # dmu2
        [0, 0, 1e-16, 0, 0, 0, 0],  # dsig2
        [0, 0, 0, 1e-16, 0, 0, 0],  # dhtail2
        [0, 0, 0, 0, 1e-02, 0, 0],  # dtau2
        [0, 0, 0, 0, 0, 1e-16, 0],  # dbg02
        [0, 0, 0, 0, 0, 0, 1e-16],  # dhs2
    ]

    amp, mu, sig, htail, tau, bg0, hstep = pars
    fwhm, dfwhm = pgb.hpge_peak_fwhm(sig, htail, tau, cov)
    assert fwhm == approx(np.log(2), rel=1e-5)
    assert dfwhm == approx(np.log(2) / 10, rel=1e-5)



import numpy as np

import pygama.math.least_squares as pgls


def test_linear_fit_by_sums():
    x = np.arange(0, 10)
    y = 2 * x + 1
    m, b = pgls.linear_fit_by_sums(x, y)

    assert m == 2
    assert b == 1


def test_fit_simple_scaling():
    x = np.arange(0, 10)
    y = 2 * x + 1
    m, b = pgls.fit_simple_scaling(x, y)

    assert m == 2.1578947368421053
    assert b == 0.0035087719298245615



import numpy as np
from scipy.stats import norm, uniform

from pygama.math.functions.gauss_on_uniform import gauss_on_uniform
from pygama.math.functions.sum_dists import SumDists


def test_gauss_on_uniform_pdf():
    x = np.arange(-10, 10)
    mu = 1
    sigma = 2
    n_sig = 10
    n_bkg = 20

    pars = np.array([np.amin(x), np.amax(x), n_sig, mu, sigma, n_bkg], dtype=float)

    assert isinstance(gauss_on_uniform, SumDists)

    y_direct = gauss_on_uniform.get_pdf(x, *pars)

    scipy_gauss = n_sig * norm.pdf(x, mu, sigma)
    scipy_uniform = n_bkg * uniform.pdf(x, np.amin(x), np.amax(x))

    scipy_y = scipy_gauss + scipy_uniform

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    y_sig, y_ext = gauss_on_uniform.pdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig + n_bkg, rtol=1e-3)


def test_gauss_on_uniform_cdf():
    x = np.arange(-10, 10)
    mu = 1
    sigma = 2
    n_sig = 10
    n_bkg = 20

    pars = np.array([np.amin(x), np.amax(x), n_sig, mu, sigma, n_bkg], dtype=float)

    assert isinstance(gauss_on_uniform, SumDists)

    y_direct = gauss_on_uniform.get_cdf(x, *pars)

    scipy_gauss = n_sig * norm.cdf(x, mu, sigma)
    scipy_uniform = n_bkg * uniform.cdf(x, np.amin(x), np.amax(x))

    scipy_y = scipy_gauss + scipy_uniform

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    y_ext = gauss_on_uniform.cdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)


def test_required_args():
    names = gauss_on_uniform.required_args()
    assert names[0] == "x_lo"
    assert names[1] == "x_hi"
    assert names[2] == "n_sig"
    assert names[3] == "mu"
    assert names[4] == "sigma"
    assert names[5] == "n_bkg"


def test_name():
    assert gauss_on_uniform.name == "gauss_on_uniform"



import pytest
from scipy.stats import norm

from pygama.math.functions.gauss import gaussian
from pygama.math.functions.sum_dists import SumDists


def test_only_two_distributions():
    with pytest.raises(ValueError) as exc_info:
        SumDists([(gaussian, [0, 1])], [0], "fracs")

    exception_raised = exc_info.value
    assert str(exception_raised) == "Can only sum two distributions at once."


def test_two_distributions_no_par_array():
    with pytest.raises(ValueError) as exc_info:
        SumDists([(gaussian, [0, 1]), (gaussian, [], [])], [0], "fracs")

    exception_raised = exc_info.value
    assert (
        str(exception_raised)
        == "Each tuple needs a distribution and a parameter index array."
    )


def test_pygama_continuous_check():
    with pytest.raises(ValueError) as exc_info:
        SumDists([(gaussian, [0, 1]), (norm, [1, 0])], [0], "fracs")

    exception_raised = exc_info.value
    assert str(exception_raised)[-12:] == "distribution"


def test_fracs_flag_with_wrong_args():
    with pytest.raises(ValueError) as exc_info:
        SumDists([(gaussian, [0, 1]), (gaussian, [1, 0])], [0, 2], "fracs")

    exception_raised = exc_info.value
    assert (
        str(exception_raised)
        == "SumDists only accepts the parameter position of one fraction."
    )


def test_areas_flag_with_wrong_args():
    with pytest.raises(ValueError) as exc_info:
        SumDists([(gaussian, [0, 1]), (gaussian, [1, 0])], [0], "areas")

    exception_raised = exc_info.value
    assert str(exception_raised) == "SumDists needs two parameter indices of areas."


def test_onea_area_flag_with_wrong_args():
    with pytest.raises(ValueError) as exc_info:
        SumDists([(gaussian, [0, 1]), (gaussian, [1, 0])], [0, 2], "one_area")

    exception_raised = exc_info.value
    assert str(exception_raised) == "SumDists needs one parameter index of an area."



from __future__ import annotations

import logging
import re
from pathlib import Path

import awkward as ak
import lgdo
from lgdo import lh5
from lgdo.types import Struct, Table, VectorOfVectors

from . import tcm as ptcm

log = logging.getLogger(__name__)


def readd_attrs(final_table, input_table):
    if hasattr(input_table, "attrs"):
        final_table.attrs = input_table.attrs
    for key in input_table:
        if hasattr(input_table[key], "attrs"):
            final_table[key].attrs = input_table[key].attrs
    if isinstance(final_table[key], (Table, Struct)):
        readd_attrs(final_table[key], input_table[key])
    if isinstance(input_table, VectorOfVectors):
        final_table[key].flattened_data.attrs = input_table[key].flattened_data.attrs
        final_table[key].cumulative_length.attrs = input_table[
            key
        ].cumulative_length.attrs


def _concat_tables(tbls):
    out_tbl = tbls[0].view_as("ak")
    for tbl in tbls[1:]:
        out_tbl = ak.concatenate([out_tbl, tbl.view_as("ak")], axis=0)
    out_tbl = Table(col_dict=out_tbl)
    readd_attrs(out_tbl, tbls[0])
    return out_tbl


def build_tcm(
    input_tables: list[tuple[str, str | list[str]]],
    coin_cols: str | list[str],
    hash_func: str | None = r"\d+",
    coin_windows: float | list[float] = 0,
    window_refs: str | list[str] = "last",
    out_file: str | None = None,
    out_name: str = "tcm",
    wo_mode: str = "write_safe",
    buffer_len: int | None = None,
    out_fields: str | list[str] | None = None,
) -> lgdo.Table | None:
    r"""Build a Time Coincidence Map (TCM).

    Given a list of input tables, create an output table containing an entry
    list of coincidences among the inputs. Uses
    :func:`.evt.tcm.generate_tcm_cols`. For use with the
    :class:`~.flow.data_loader.DataLoader`.

    Parameters
    ----------
    input_tables
        Each entry is ``(filename, table_name_pattern)``. ``table_name_pattern``
        may be a string or list of strings. All tables matching each pattern in
        ``filename`` will be used as input tables.
    coin_cols
        Name of the column (or columns) in each table used to build
        coincidences. All input tables must contain these columns.
    hash_func
        mapping of table names to integers for use in the TCM.  `hash_func` is
        a regexp pattern that acts on each table name. The default `hash_func`
        ``r"\d+"`` pulls the first integer out of the table name. Setting to
        ``None`` will use a table's index in `input_tables`.
    coin_windows
        Width of the clustering window(s). If a single value is supplied it will
        be used for all ``coin_cols``.
    window_refs
        Window reference for the clustering window. Currently only ``"last"`` is
        implemented.
    out_file
        name (including path) for the output file. If ``None``, no file will be
        written; the TCM will just be returned in memory.
    out_name
        name for the TCM table in the output file.
    wo_mode
        mode to send to :meth:`~.lgdo.lh5.LH5Store.write`.

    out_fields
        Optional additional fields to propagate from the input tables into the
        output TCM.

    Returns
    -------
    lgdo.Table or None
        If ``out_file`` is ``None`` the resulting TCM is returned as a
        :class:`lgdo.Table`. Otherwise ``None`` is returned after writing the
        table to ``out_file``.

    See Also
    --------
    .tcm.generate_tcm_cols
    """
    # hash_func: later can add list or dict or a function(str) --> int.

    if not isinstance(coin_cols, list):
        coin_cols = [coin_cols]
    if not isinstance(coin_windows, list):
        coin_windows = [coin_windows]
    if not isinstance(window_refs, list):
        window_refs = [window_refs]
    if out_fields is not None and not isinstance(out_fields, list):
        out_fields = [out_fields]

    if len(coin_cols) != len(coin_windows):
        if len(coin_windows) == 1:
            coin_windows = coin_windows * len(coin_cols)
        else:
            msg = (
                "coin_cols and coin_windows must have the same length, "
                f"got {len(coin_cols)} and {len(coin_windows)}"
            )
            raise ValueError(msg)
    if len(coin_cols) != len(window_refs):
        if len(window_refs) == 1:
            window_refs = window_refs * len(coin_cols)
        else:
            msg = (
                "coin_cols and coin_windows must have the same length, "
                f"got {len(coin_cols)} and {len(window_refs)}"
            )
            raise ValueError(msg)

    iterators = []
    table_keys = []
    all_tables = []

    if buffer_len is None:
        ntables = 0
        for filename, patterns in input_tables:
            if isinstance(patterns, str):
                patterns = [patterns]
            for pattern in patterns:
                ntables += len(lh5.ls(filename, lh5_group=pattern))

        n_fields = (
            2 + len(set(coin_cols + out_fields))
            if out_fields is not None
            else 2 + len(set(coin_cols))
        )
        buffer_len = int(10**7 / (ntables * n_fields))

    for filename, patterns in input_tables:
        if isinstance(patterns, str):
            patterns = [patterns]
        for pattern in patterns:
            tables = lh5.ls(filename, lh5_group=pattern)
            for table in tables:
                all_tables.append(table)
                table_key = len(table_keys)
                if hash_func is not None:
                    if isinstance(hash_func, str):
                        table_key = int(re.search(hash_func, table).group())
                    else:
                        raise NotImplementedError(
                            f"hash_func of type {type(hash_func).__name__}"
                        )
                else:
                    table_key = len(all_tables) - 1

                h5py_open_mode = "a" if out_file == filename else "r"
                iterators.append(
                    lh5.LH5Iterator(
                        filename,
                        table,
                        field_mask=coin_cols,
                        buffer_len=buffer_len,
                        h5py_open_mode=h5py_open_mode,
                    )
                )
                table_keys.append(table_key)

    coin_windows = [
        ptcm.coin_groups(n, w, r)
        for n, w, r in zip(coin_cols, coin_windows, window_refs)
    ]

    tcm_gen = ptcm.generate_tcm_cols(
        iterators, coin_windows=coin_windows, table_keys=table_keys, fields=out_fields
    )
    tcm = []
    # clear existing output files
    if out_file is not None and wo_mode == "of":
        if Path(out_file).exists():
            Path(out_file).unlink()
    while True:
        try:
            out_tbl = tcm_gen.__next__()
            out_tbl.attrs.update(
                {"tables": str(all_tables), "hash_func": str(hash_func)}
            )
            if out_file is not None:
                lh5.write(
                    out_tbl, out_name, out_file, wo_mode="o" if wo_mode == "u" else "a"
                )
            else:
                tcm.append(out_tbl)
        except StopIteration:
            break
    if out_file is None:
        out_tbl = _concat_tables(tcm)
        return out_tbl



r"""
Provide a convenience function for a Gaussian on top of an extended Gaussian. The correct usage is :func:`gauss_on_exgauss.get_pdf(x, mu, sigma, htail, tau)`

Parameters
----------
mu
    The location and scale of the first Gaussian
htail
    The fraction of the Gaussian tail vs. the Gaussian peak
tau
    The characteristic scale of the extended Gaussian tail

Returns
-------
gauss_on_exgauss
    A subclass of :class:`SumDists` and :class:`rv_continuous`, has methods of :func:`pdf`, :func:`cdf`, etc.

Notes
-----
The extended Gaussian distribution shares the mu, sigma with the Gaussian
"""

from pygama.math.functions.exgauss import exgauss
from pygama.math.functions.gauss import gaussian
from pygama.math.functions.sum_dists import SumDists

(mu, sigma, frac, tau) = range(
    4
)  # the sum_dist array we will pass will be mu, sigma, frac/htail, tau
par_array = [(exgauss, [mu, sigma, tau]), (gaussian, [mu, sigma])]

gauss_on_exgauss = SumDists(
    par_array,
    [frac],
    "fracs",
    parameter_names=["mu", "sigma", "htail", "tau"],
    name="gauss_on_exgauss",
)



r"""
Provide a convenience function for a Gaussian on top of a step distribution.

Parameters
----------
x_lo, x_hi
    The lower and upper bounds of the support of the step function
area1
    The area of the Gaussian distribution
mu, sigma
    The location and scale of the first Gaussian, and step function
area2
    The area of the step function
hstep
    The height of the step function

Returns
-------
gauss_on_step
    An instance of SumDists and rv_continuous, has methods of pdf, cdf, etc.

Notes
-----
The step function shares a mu and sigma with the step function
The parameter array must have ordering (x_lo, x_hi, area1, mu, sigma, area2, hstep)
"""

from pygama.math.functions.gauss import gaussian
from pygama.math.functions.step import step
from pygama.math.functions.sum_dists import SumDists

(x_lo, x_hi, n_sig, mu, sigma, n_bkg, hstep) = range(
    7
)  # the sum_dist array we will pass will be x_lo, x_hi, n_sig, mu, sigma, n_bkg, hstep
par_array = [(gaussian, [mu, sigma]), (step, [x_lo, x_hi, mu, sigma, hstep])]

gauss_on_step = SumDists(
    par_array,
    [n_sig, n_bkg],
    "areas",
    parameter_names=["x_lo", "x_hi", "n_sig", "mu", "sigma", "n_bkg", "hstep"],
    name="gauss_on_step",
)



"""
This module implements routines to evaluate expressions to columnar data.
"""

from __future__ import annotations

import logging
import os
from collections import OrderedDict
from typing import Iterable, Mapping

import lgdo
import lgdo.lh5 as lh5
import numpy as np
from lgdo.lh5 import LH5Iterator, ls

from .. import utils

log = logging.getLogger(__name__)


def build_hit(
    infile: str,
    outfile: str = None,
    hit_config: str | Mapping = None,
    lh5_tables: Iterable[str] = None,
    lh5_tables_config: str | Mapping[str, Mapping] = None,
    n_max: int = np.inf,
    wo_mode: str = "write_safe",
    buffer_len: int = 3200,
) -> None:
    """
    Transform a :class:`~lgdo.types.table.Table` into a new
    :class:`~lgdo.types.table.Table` by evaluating strings describing column
    operations.

    Operates on columns only, not specific rows or elements. Relies on
    :meth:`~lgdo.types.table.Table.eval`.

    Parameters
    ----------
    infile
        input LH5 file name containing tables to be processed.
    outfile
        name of the output LH5 file. If ``None``, create a file in the same
        directory and append `_hit` to its name.
    hit_config
        dictionary or name of JSON file defining column transformations. Must
        contain an ``outputs`` and an ``operations`` keys. For example:

        .. code-block:: json

            {
              "outputs": ["calE", "AoE"],
              "operations": {
                "calE": {
                  "expression": "sqrt(a + b * trapEmax**2)",
                  "parameters": {"a": "1.23", "b": "42.69"},
                },
                "AoE": {"expression": "A_max/calE"},
              }
            }

        The ``outputs`` array lists columns that will be effectively written in
        the output LH5 file. Add here columns that will be simply forwarded as
        they are from the DSP tier.

    lh5_tables
        tables to consider in the input file. if ``None``, tables with name
        `dsp` will be searched for in the file, even nested by one level.
    lh5_tables_config
        dictionary or JSON file defining the mapping between LH5 tables in
        `infile` and hit configuration. Table names can be directly mapped to
        configuration blocks or to JSON files containing them. This option is
        mutually exclusive with `hit_config` and `lh5_tables`.
    n_max
        maximum number of rows to process
    wo_mode
        forwarded to :meth:`lgdo.lh5.write`.

    See Also
    --------
    lgdo.types.table.Table.eval
    """

    if lh5_tables_config is None and hit_config is None:
        raise ValueError("either lh5_tables_config or hit_config must be specified")

    if lh5_tables_config is not None and (
        hit_config is not None or lh5_tables is not None
    ):
        raise ValueError(
            "lh5_tables_config and hit_config/lh5_tables options are mutually exclusive"
        )

    if lh5_tables_config is not None:
        tbl_cfg = lh5_tables_config
        # sanitize config
        if isinstance(tbl_cfg, str):
            tbl_cfg = utils.load_dict(tbl_cfg)

        for k, v in tbl_cfg.items():
            if isinstance(v, str):
                tbl_cfg[k] = utils.load_dict(v)
        lh5_tables_config = tbl_cfg

    else:
        if isinstance(hit_config, str):
            # sanitize config
            hit_config = utils.load_dict(hit_config)

        lh5_tables_config = {}
        if lh5_tables is None:
            if "dsp" in ls(infile):
                log.debug("found candidate table /dsp")
                lh5_tables_config["dsp"] = hit_config
            for el in ls(infile):
                if f"{el}/dsp" in ls(infile, f"{el}/"):
                    log.debug(f"found candidate table /{el}/dsp")
                    lh5_tables_config[f"{el}/dsp"] = hit_config
        else:
            for tbl in lh5_tables:
                lh5_tables_config[tbl] = hit_config

    if outfile is None:
        outfile = os.path.splitext(os.path.basename(infile))[0]
        outfile = outfile.removesuffix("_dsp") + "_hit.lh5"

    # reorder blocks in "operations" based on dependency
    log.debug("reordering operations based on mutual dependency")
    for cfg in lh5_tables_config.values():
        cfg["operations"] = _reorder_table_operations(cfg["operations"])

    first_done = False
    for tbl, cfg in lh5_tables_config.items():
        lh5_it = LH5Iterator(infile, tbl, buffer_len=buffer_len)
        write_offset = 0

        log.info(f"Processing table '{tbl}' in file {infile}")

        for tbl_obj in lh5_it:
            start_row = lh5_it.current_i_entry

            # create a new table object that links all the columns in the
            # current table (i.e. no copy)
            outtbl_obj = lgdo.Table(col_dict=tbl_obj)

            for outname, info in cfg["operations"].items():
                outcol = outtbl_obj.eval(
                    info["expression"], info.get("parameters", None)
                )
                if "lgdo_attrs" in info:
                    outcol.attrs |= info["lgdo_attrs"]

                log.debug(f"made new column {outname!r}={outcol!r}")
                outtbl_obj.add_column(outname, outcol)

            # make high level flags
            if "aggregations" in cfg:
                for high_lvl_flag, flags in cfg["aggregations"].items():
                    flags_list = list(flags.values())
                    n_flags = len(flags_list)
                    if n_flags <= 8:
                        flag_dtype = np.uint8
                    elif n_flags <= 16:
                        flag_dtype = np.uint16
                    elif n_flags <= 32:
                        flag_dtype = np.uint32
                    else:
                        flag_dtype = np.uint64

                    df_flags = outtbl_obj.view_as("pd", cols=flags_list)
                    flag_values = df_flags.values.astype(flag_dtype)

                    multiplier = 2 ** np.arange(n_flags, dtype=flag_values.dtype)
                    flag_out = np.dot(flag_values, multiplier)

                    outtbl_obj.add_field(high_lvl_flag, lgdo.Array(flag_out))

            # remove or add columns according to "outputs" in the configuration
            # dictionary
            if "outputs" in cfg:
                if isinstance(cfg["outputs"], list):
                    # add missing columns (forwarding)
                    for out in cfg["outputs"]:
                        if out not in outtbl_obj:
                            outtbl_obj.add_column(out, tbl_obj[out])

                    # remove non-required columns
                    existing_cols = list(outtbl_obj.keys())
                    for col in existing_cols:
                        if col not in cfg["outputs"]:
                            outtbl_obj.remove_column(col, delete=True)

            lh5.write(
                obj=outtbl_obj,
                name=tbl.replace("/dsp", "/hit"),
                lh5_file=outfile,
                n_rows=len(tbl_obj),
                wo_mode=wo_mode if first_done is False else "append",
                write_start=write_offset + start_row,
            )

            first_done = True


def _reorder_table_operations(
    config: Mapping[str, Mapping],
) -> OrderedDict[str, Mapping]:
    """Reorder operations in `config` according to mutual dependency."""

    def _one_pass(config):
        """Loop once over `config` and do a first round of reordering"""
        # list to hold reordered config keys
        ordered_keys = []

        # start looping over config
        for outname in config:
            # initialization
            if not ordered_keys:
                ordered_keys.append(outname)
                continue

            if outname in ordered_keys:
                raise RuntimeError(f"duplicated operation '{outname}' detected")

            # loop over existing reordered keys and figure out where to place
            # the new key
            idx = 0
            for k in ordered_keys:
                # get valid names in the expression
                c = compile(
                    config[k]["expression"], "gcc -O3 -ffast-math build_hit.py", "eval"
                )

                # if we need "outname" for this expression, insert it before!
                if outname in c.co_names:
                    break
                else:
                    idx += 1

            ordered_keys.insert(idx, outname)

        # now replay the config dictionary based on sorted keys
        opdict = OrderedDict()
        for k in ordered_keys:
            opdict[k] = config[k]

        return opdict

    # okay, now we need to repeat this until we've sorted everything
    current = OrderedDict(config)

    while True:
        new = _one_pass(current)

        if new == current:
            return new
        else:
            current = new


def _get_dependencies(config, par, pars=None):
    """
    Recursive func to iterate back through tree of input blocks for a given output block
    """
    if pars is None:
        pars = []
    par_op = config[par]
    c = compile(par_op["expression"], "gcc -O3 -ffast-math build_hit.py", "eval")
    for p in c.co_names:
        if p in par_op["parameters"]:
            pass
        else:
            pars.append(p)
            if p in config:
                pars = _get_dependencies(config, p, pars)
    return pars


def _remove_uneeded_operations(config, outpars):
    """
    Function that removes any operations not needed to generate outpars from the config dictionary
    Returns the config without these blocks as well as a list of input keys from the dsp file
    needed to generate outpars
    """
    if not isinstance(outpars, list):
        outpars = [outpars]
    dependent_keys = [*outpars]
    inkeys = []
    for par in outpars:
        pars = _get_dependencies(config, par)
        for p in pars:
            if p in config and p not in dependent_keys:
                dependent_keys.append(p)
            elif p not in config and p not in inkeys:
                inkeys.append(p)

    for key in list(config):
        if key not in dependent_keys:
            config.pop(key)
    return config, inkeys
